import logging
import os

from typing_extensions import override

from bpm_ai_core.translation.nmt import NMTModel

try:
    from azure.ai.translation.text import TranslatorCredential
    from azure.ai.translation.text.aio import TextTranslationClient
    from azure.ai.translation.text.models import InputTextItem

    has_azure_translation = True
except ImportError:
    has_azure_translation = False

azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.WARNING)


class AzureTranslation(NMTModel):
    """Azure Text Translation NMT Model"""

    def __init__(self, endpoint: str = None, region: str = "westeurope"):
        if not has_azure_translation:
            raise ImportError('azure-ai-translation-text  is not installed')
        self.key = os.getenv('AZURE_TRANSLATION_KEY')
        self.endpoint = endpoint or os.getenv('AZURE_TRANSLATION_ENDPOINT')
        self.region = region or os.getenv('AZURE_TRANSLATION_REGION')

    @override
    async def _do_translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        async with TextTranslationClient(
                endpoint=self.endpoint,
                credential=TranslatorCredential(self.key, self.region)
        ) as client:
            if isinstance(text, str):
                texts = [text]
            else:
                texts = text
            items = [InputTextItem(text=text) for text in texts]
            response = await client.translate(content=items, to=[target_language])
            return [doc.translations[0].text for doc in response]



import os
from urllib.parse import urlparse

try:
    from aiobotocore.session import get_session
except ImportError:
    pass

try:
    from azure.storage.blob.aio import BlobClient
except ImportError:
    pass


def is_s3_url(url: str) -> bool:
    return url.startswith('s3://') or (url.startswith('https://') and "amazonaws.com" in url)


async def read_file_from_s3(file_url: str) -> bytes:
    """
    Reads a file from an S3 bucket using a URL-like format.

    Args:
        file_url (str): The URL-like string representing the S3 file location.
                        Formats:
                          - "s3://<bucket-name>/<file-path>"
                          - "https://<bucket-name>.s3.<region>.amazonaws.com/<file-path>"

    Returns:
        bytes: The contents of the file.
    """
    try:
        bucket_name, file_path = await parse_s3_url(file_url)

        async with get_session().create_client('s3') as s3:
            response = await s3.get_object(Bucket=bucket_name, Key=file_path)
            async with response['Body'] as stream:
                return await stream.read()
    except Exception as e:
        raise Exception(f"Error reading file from S3: {str(e)}")


async def parse_s3_url(s3_url: str):
    # Extract bucket name and file path based on the URL format
    parsed_url = urlparse(s3_url)
    if parsed_url.scheme == 's3':
        bucket_name = parsed_url.netloc
        file_path = parsed_url.path.lstrip('/')
    elif parsed_url.scheme == 'https' and '.amazonaws.com' in parsed_url.netloc:
        bucket_name = parsed_url.netloc.split('.')[0]
        file_path = parsed_url.path.lstrip('/')
    else:
        raise ValueError("Invalid S3 file URL format.")
    return bucket_name, file_path


def is_azure_blob_url(url: str) -> bool:
    return "blob.core.windows.net" in url


async def read_file_from_azure_blob(file_url: str) -> bytes:
    """
       Reads a file from Azure Blob Storage using a URL.

       Args:
           file_url (str): The URL of the file in Azure Blob Storage.
                           Format: "https://<account-name>.blob.core.windows.net/<container-name>/<file-path>"

       Returns:
           str: The contents of the file as a string.
       """
    try:
        access_key = os.environ.get('AZURE_STORAGE_ACCESS_KEY')
        async with BlobClient.from_blob_url(file_url, credential=access_key) as blob_client:
            blob = await blob_client.download_blob()
            return await blob.readall()
    except Exception as e:
        raise Exception(f"Error reading file from Azure Blob Storage: {str(e)}")



import xmltodict


def dict_to_xml(data, root_tag='root'):
    """
    Serializes a dict to XML.

    Args:
        data (dict): The dict to be serialized to XML.
        root_tag (str): The name of the root tag in the XML output. Default is 'root'.

    Returns:
        str: The serialized XML string.
    """
    return xmltodict.unparse({root_tag: data}, pretty=True, full_document=False)


def xml_to_dict(xml_data):
    """
    Deserializes XML to a dict.

    Args:
        xml_data (str): The XML string to be deserialized.

    Returns:
        dict: The deserialized dict.
    """
    data_dict = xmltodict.parse(xml_data)
    return next(iter(data_dict.values()))



import io
from abc import ABC, abstractmethod
from typing import Optional, Union

from pydantic import BaseModel

from bpm_ai_core.tracing.decorators import span
from bpm_ai_core.util.audio import load_audio

class ASRResult(BaseModel):
    text: str


class ASRModel(ABC):
    """
    Automatic Speech Recognition (ASR) model for transcribing audio.
    """

    @abstractmethod
    async def _do_transcribe(self, audio: io.BytesIO, language: Optional[str] = None) -> ASRResult:
        pass

    @span(name="asr")
    async def transcribe(self, audio_or_path: Union[io.BytesIO, str], language: Optional[str] = None) -> ASRResult:
        if isinstance(audio_or_path, str):
            audio = load_audio(audio_or_path)
        else:
            audio = audio_or_path
        return await self._do_transcribe(audio, language)



from bpm_ai_core.tracing.tracer import Tracer


class DelegateTracer(Tracer):
    def __init__(self, tracers: list[Tracer]):
        self.tracers = tracers

    def start_trace(self, name, inputs, tags=None):
        for tracer in self.tracers:
            tracer.start_trace(name, inputs, tags)

    def end_trace(self, outputs, error_msg=None):
        for tracer in self.tracers:
            tracer.end_trace(outputs)

    def start_span(self, name, inputs):
        for tracer in self.tracers:
            tracer.start_span(name, inputs)

    def end_span(self, outputs, error_msg=None):
        for tracer in self.tracers:
            tracer.end_span(outputs, error_msg)

    def start_llm_trace(self, llm, messages, current_try, tools=None):
        for tracer in self.tracers:
            tracer.start_llm_trace(llm, messages, current_try, tools)

    def end_llm_trace(self, completion=None, error_msg=None):
        for tracer in self.tracers:
            tracer.end_llm_trace(completion, error_msg)

    def start_tool_trace(self, tool, inputs):
        for tracer in self.tracers:
            tracer.start_tool_trace(tool, inputs)

    def end_tool_trace(self, output=None, error_msg=None):
        for tracer in self.tracers:
            tracer.end_tool_trace(output, error_msg)

    def event(self, name: str, inputs: dict | None = None, outputs: dict | None = None, error_msg: str | None = None):
        for tracer in self.tracers:
            tracer.event(name, inputs, outputs, error_msg)

    def finalize(self):
        for tracer in self.tracers:
            tracer.finalize()



import json

from bpm_ai_core.util.markdown import dict_to_md
from bpm_ai_core.util.xml_convert import dict_to_xml as _dict_to_xml


def dict_to_markdown(d: dict) -> str:
    return dict_to_md(d).strip()


def dict_to_json(d: dict) -> str:
    return json.dumps(d, indent=2)


def dict_to_xml(d: dict, root: str = "root") -> str:
    return _dict_to_xml(d, root_tag=root)



from abc import ABC, abstractmethod
from typing import Tuple

from pydantic import BaseModel, Field

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.tracing.decorators import span


class OCRPage(BaseModel):
    text: str
    words: list[str] = Field(..., exclude=True)
    bboxes: list[Tuple[float, float, float, float]] = Field(..., exclude=True)
    """Format: (x, y, x + w, y + h), normalized to 1"""


class OCRResult(BaseModel):
    pages: list[OCRPage]

    @property
    def full_text(self) -> str:
        return "\n".join([p.text for p in self.pages])


class OCR(ABC):
    """
    Optical Character Recognition (OCR) Model
    """

    @abstractmethod
    async def _do_process(
            self,
            blob: Blob,
            language: str = None
    ) -> OCRResult:
        pass

    @span(name="ocr")
    async def process(
            self,
            blob_or_path: Blob | str,
            language: str = None
    ) -> OCRResult:
        if isinstance(blob_or_path, str):
            blob = Blob.from_path_or_url(blob_or_path)
        else:  # Blob
            blob = blob_or_path
        return await self._do_process(
            blob=blob,
            language=language
        )



from abc import ABCMeta, abstractmethod


class Tracer(metaclass=ABCMeta):
    @abstractmethod
    def start_trace(self, name: str, inputs: dict, tags: list[str] = None):
        pass

    @abstractmethod
    def end_trace(self, outputs: dict, error_msg: str | None = None):
        pass

    @abstractmethod
    def start_span(self, name: str, inputs: dict):
        pass

    @abstractmethod
    def end_span(self, outputs: dict, error_msg: str | None = None):
        pass

    @abstractmethod
    def start_llm_trace(self, llm, messages, current_try, tools=None):
        pass

    @abstractmethod
    def end_llm_trace(self, completion=None, error_msg=None):
        pass

    @abstractmethod
    def start_tool_trace(self, tool, inputs: dict):
        pass

    @abstractmethod
    def end_tool_trace(self, outputs: dict | None = None, error_msg: str | None = None):
        pass

    @abstractmethod
    def event(self, name: str, inputs: dict | None = None, outputs: dict | None = None, error_msg: str | None = None):
        pass

    @abstractmethod
    def finalize(self):
        pass



import hashlib
import logging
from contextvars import ContextVar

from bpm_ai_core.llm.openai_chat._constants import default_client_kwargs, AZURE_API_KEY_ENV_VAR

logger = logging.getLogger(__name__)

try:
    from openai import AsyncOpenAI
    from openai.lib.azure import AsyncAzureOpenAI

    _clients: ContextVar[dict[str, AsyncOpenAI]] = ContextVar('openai_clients', default={})
except ImportError:
    pass


def get_openai_client(endpoint: str = None, api_key: str = None) -> AsyncOpenAI:
    client_map = _clients.get()
    hash_key = hashlib.sha256(((endpoint or "default") + (api_key or "default")).encode()).hexdigest()
    if hash_key in client_map.keys():
        return client_map[hash_key]
    else:
        client = AsyncOpenAI(
            base_url=endpoint,
            api_key=api_key,
            **default_client_kwargs
        )
        client_map[hash_key] = client
        return client


def get_azure_openai_client(azure_endpoint: str, api_version: str, api_key: str) -> AsyncOpenAI:
    client_map = _clients.get()
    hash_key = hashlib.sha256((azure_endpoint + api_key).encode()).hexdigest()
    if hash_key in client_map.keys():
        return client_map[hash_key]
    else:
        client = AsyncAzureOpenAI(
            azure_endpoint=azure_endpoint,
            api_key=api_key,
            api_version=api_version,
            **default_client_kwargs
        )
        client_map[hash_key] = client
        return client



import logging
import os
from io import BytesIO

from typing_extensions import override

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.ocr.ocr import OCR, OCRResult, OCRPage
from bpm_ai_core.util.image import blob_as_images

try:
    from azure.ai.documentintelligence.aio import DocumentIntelligenceClient as AsyncDocumentIntelligenceClient
    from azure.core.credentials import AzureKeyCredential

    has_azure_doc = True
except ImportError:
    has_azure_doc = False

azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.WARNING)

IMAGE_FORMATS = ["png", "jpeg", "tiff"]


class AzureOCR(OCR):
    def __init__(self, endpoint: str = None):
        if not has_azure_doc:
            raise ImportError('azure-ai-documentintelligence is not installed')
        self.endpoint = endpoint or os.environ.get("AZURE_DOCUMENT_INTELLIGENCE_ENDPOINT")

    @override
    async def _do_process(
        self,
        blob: Blob,
        language: str = None
    ) -> OCRResult:
        if not (blob.is_pdf() or blob.is_image()):
            raise ValueError("Blob must be a PDF or an image")
        if blob.is_pdf():
            bytes_io = await blob.as_bytes_io()
        else:
            bytes_io = BytesIO((await blob_as_images(blob, accept_formats=IMAGE_FORMATS, return_bytes=True))[0])

        async with AsyncDocumentIntelligenceClient(
                self.endpoint,
                AzureKeyCredential(
                    os.environ.get("AZURE_DOCUMENT_INTELLIGENCE_KEY")
                )
        ) as client:
            document = await client.begin_analyze_document(
                model_id="prebuilt-layout",
                analyze_request=bytes_io,
                content_type="application/octet-stream",
                output_content_format="markdown"
            )

            # Wait for the extraction to complete asynchronously
            result = await document.result()
            markdown_content = result.content

            pages = []
            for page in result.pages:
                bboxes = []
                words = []
                for word in page.words:
                    polygon = word.polygon
                    x, y = polygon[0], polygon[1]
                    w, h = polygon[2] - x, polygon[5] - y
                    bboxes.append((x / page['width'], y / page['height'], (x + w) / page['width'], (y + h) / page['height']))
                    words.append(word.content)

                page_data = OCRPage(
                    text=" ".join(words),
                    words=words,
                    bboxes=bboxes
                )
                pages.append(page_data)

        return OCRResult(pages=pages)



from inspect import signature
from typing import Any, Optional, Dict, Callable, Type, Union

from pydantic import BaseModel, validate_call, create_model

from bpm_ai_core.util.json_schema import expand_simplified_json_schema


def _create_subset_model(
    name: str, model: BaseModel, field_names: list
) -> Type[BaseModel]:
    """Create a pydantic model with only a subset of model's fields."""
    fields = {}
    for field_name in field_names:
        field = model.model_fields[field_name]
        fields[field_name] = (field.annotation.__origin__, field)
    return create_model(name, **fields)


def _get_filtered_args(
    inferred_model: Type[BaseModel],
    func: Callable,
) -> dict:
    """Get the arguments from a function's signature."""
    schema = inferred_model.model_json_schema()["properties"]
    valid_keys = signature(func).parameters
    return {k: schema[k] for k in valid_keys}


class _SchemaConfig:
    """Configuration for the pydantic model."""
    arbitrary_types_allowed: bool = True


def create_schema_from_function(
    model_name: str,
    func: Callable,
) -> Type[BaseModel]:
    """Create a pydantic schema from a function's signature.
    Args:
        model_name: Name to assign to the generated pydandic schema
        func: Function to generate the schema from
    Returns:
        A pydantic model with the same arguments as the function
    """
    validated = validate_call(func, config=_SchemaConfig)  # type: ignore
    inferred_model = validated.__pydantic_core_schema__
    # Pydantic adds placeholder virtual fields we need to strip
    valid_properties = _get_filtered_args(inferred_model, func)
    return _create_subset_model(
        f"{model_name}Schema", inferred_model, list(valid_properties)
    )


class Tool(BaseModel):

    name: str
    """
    The unique name of the tool that clearly communicates its purpose.
    """

    description: str
    """
    A description of what the function does, used by the model to choose when and how to use the tool.
    """

    args_schema: Optional[Dict[str, Any]] = None
    """Function's input argument schema."""

    callable: Optional[Callable] = None

    @classmethod
    def create(
        cls,
        name: str,
        description: str,
        args_schema: Union[Dict[str, Any], Type[BaseModel]],
        callable: Optional[Callable] = None
    ):
        if isinstance(args_schema, dict):
            args_schema = expand_simplified_json_schema(args_schema)
        else:
            json_schema = args_schema.model_json_schema()
            json_schema.pop("title")
            args_schema = json_schema
        return cls(
           name=name,
            description=description,
            args_schema=args_schema,
            callable=callable
        )





from abc import ABC, abstractmethod

from pydantic import BaseModel

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.tracing.decorators import span


class QAResult(BaseModel):
    answer: str | None
    score: float
    start_index: int | None
    end_index: int | None


class QuestionAnswering(ABC):
    """
    (Extractive) Question Answering Model
    """

    @abstractmethod
    async def _do_answer(
            self,
            context_str_or_blob: str | Blob,
            question: str
    ) -> QAResult:
        pass

    @span(name="qa")
    async def answer(
            self,
            context_str_or_blob: str | Blob,
            question: str,
            confidence_threshold: float | None = 0.1
    ) -> QAResult | None:
        result = await self._do_answer(
            context_str_or_blob=context_str_or_blob,
            question=question
        )
        # Only return the answer if the score is above the threshold (if given)
        return result \
            if not confidence_threshold or result.score > confidence_threshold \
            else None



import base64
import io
import logging
import re
import tempfile
from io import BytesIO
from typing import Union, Tuple

from PIL import Image, ImageDraw
from pdf2image import convert_from_path, convert_from_bytes

logger = logging.getLogger(__name__)


image_ext_map = {
    'bmp': 'image/bmp',
    'gif': 'image/gif',
    'icns': 'image/x-icns',
    'ico': 'image/x-icon',
    'jfif': 'image/jpeg',
    'jpe': 'image/jpeg',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpeg',
    'png': 'image/png',
    'pbm': 'image/x-portable-bitmap',
    'pgm': 'image/x-portable-graymap',
    'pnm': 'image/x-portable-anymap',
    'ppm': 'image/x-portable-pixmap',
    'tif': 'image/tiff',
    'tiff': 'image/tiff',
    'webp': 'image/webp',
}

pdf_ext_map = {
    'pdf': 'application/pdf'
}


async def blob_as_images(blob, accept_formats: list[str], return_bytes: bool = False) -> Union[list[Image.Image], list[bytes]]:
    """
    Load an image, PDF, or other file in a Blob into a Pillow Image object or raw bytes of accepted format.

    Args:
        blob: The input Blob object containing the image data.
        accept_formats: A list of accepted image formats (e.g., ['png', 'jpeg']).
        return_bytes: If True, return the image data as bytes instead of PIL Image objects.

    Returns:
        A list of PIL Image objects or a list of bytes representing the converted images.
    """
    if blob.is_pdf():
        # Convert PDF to a list of images
        logger.info("Converting PDF to a list of images...")
        images = pdf_to_images(await blob.as_bytes())
    elif blob.is_image():
        # Load the image from the blob
        images = [Image.open(await blob.as_bytes_io())]
    else:
        raise ValueError(f"Unsupported blob type: {blob.mimetype}")

    accept_formats = [f.lower() for f in accept_formats]

    # Convert images to the accepted formats if necessary
    converted_images = []
    for image in images:
        if image.format.lower() not in accept_formats:
            # Convert the image to the first accepted format
            output_format = accept_formats[0]
            logger.info(f"Converting images from {image.format.lower()} to accepted format: {output_format}")
            if return_bytes:
                # Convert the image to bytes
                with io.BytesIO() as output_bytes:
                    if not image.mode == "RGB":
                        image = image.convert("RGB")
                    image.save(output_bytes, format=output_format)
                    converted_images.append(output_bytes.getvalue())
            else:
                # Convert the image to PIL Image object
                converted_image = io.BytesIO()
                if not image.mode == "RGB":
                    image = image.convert("RGB")
                image.save(converted_image, format=output_format)
                converted_image.seek(0)
                converted_images.append(Image.open(converted_image))
        else:
            if return_bytes:
                # Convert the image to bytes
                with io.BytesIO() as output_bytes:
                    image.save(output_bytes, format=image.format)
                    converted_images.append(output_bytes.getvalue())
            else:
                converted_images.append(image)

    return converted_images


def pdf_to_images(pdf: bytes | str) -> list[Image]:
    with tempfile.TemporaryDirectory() as path:
        if isinstance(pdf, bytes):
            func = convert_from_bytes
        else:
            func = convert_from_path
        return func(pdf, output_folder=path, dpi=100, use_pdftocairo=True)


def base64_encode_image(image: Image):
    """
    Get a base64 encoded string from a Pillow Image object.

    Parameters:
    - image (Image): A Pillow Image object.

    Returns:
    - str: Base64 encoded string of the image.
    """
    buffered = BytesIO()
    image.save(buffered, format=image.format or "JPEG")  # Assuming JPEG if format is not provided
    img_str = base64.b64encode(buffered.getvalue())
    return img_str.decode('utf-8')


def replace_base64_values(data):
    if isinstance(data, dict):
        decoded_dict = {}
        for key, value in data.items():
            decoded_dict[key] = replace_base64_values(value)
        return decoded_dict
    elif isinstance(data, list):
        decoded_list = []
        for item in data:
            decoded_list.append(replace_base64_values(item))
        return decoded_list
    elif isinstance(data, str) and len(data) > 1024 and is_base64(data):
        return "..."
    else:
        return data


def is_base64(s: str):
    try:
        s = re.sub(r"^data:image/.+;base64,", "", s)
        base64.b64decode(s, validate=True)
        return True
    except:
        return False


def draw_boxes_on_image(image: Image, normalized_boxes: list[Tuple[float, float, float, float]]):
    """
    Draws bounding boxes on a given PIL image and displays the resulting image.

    Args:
        image (PIL.Image): The input image.
        normalized_boxes (list): A list of normalized bounding box coordinates.
                                 Each box is represented as a tuple (x1, y1, x2, y2),
                                 where (x1, y1) is the top-left corner and (x2, y2) is the bottom-right corner.
                                 The coordinates are normalized, ranging from 0 to 1.

    Returns:
        None
    """
    draw = ImageDraw.Draw(image)
    width, height = image.size
    # Iterate over the normalized bounding boxes
    for box in normalized_boxes:
        x1, y1, x2, y2 = box
        # Convert normalized coordinates to pixel coordinates
        x1 = int(x1 * width)
        y1 = int(y1 * height)
        x2 = int(x2 * width)
        y2 = int(y2 * height)
        # Draw the bounding box rectangle
        draw.rectangle(((x1, y1), (x2, y2)), outline="red", width=2)
    # Display the image with bounding boxes
    image.show()


import inspect
import os
import re
from typing import List, Dict, Any

from jinja2 import Template, environment

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.llm.common.message import ChatMessage, AssistantMessage, ToolResultMessage, ToolCallMessage, \
    UserMessage, SystemMessage
from bpm_ai_core.prompt.filters import dict_to_markdown, dict_to_xml, dict_to_json

environment.DEFAULT_FILTERS['markdown'] = dict_to_markdown
environment.DEFAULT_FILTERS['json'] = dict_to_json
environment.DEFAULT_FILTERS['xml'] = dict_to_xml


class Prompt:

    def __init__(self, kwargs: Dict[str, Any], path: str | None = None, template_str: str | None = None) -> None:
        self.path = path
        self.template_str = template_str
        self.template_vars = kwargs

    @classmethod
    def from_file(cls, path: str, **kwargs):
        # Get the frame of the caller of the function that called this function
        caller_frame = inspect.stack()[1]
        caller_filename = caller_frame.filename

        # Get the directory of the caller's file
        current_dir = os.path.dirname(os.path.abspath(caller_filename))
        file_path = os.path.join(current_dir, path)

        return cls(kwargs, path=file_path)

    @classmethod
    def from_string(cls, template: str, **kwargs):
        return cls(kwargs, template_str=template)

    def format(self, llm_name: str = "") -> List[ChatMessage]:
        template = self.load_template(self.path, llm_name) if self.path else Template(self.template_str)
        full_prompt = template.render(self.template_vars)

        regex = r'\[#\s*(user|assistant|system|tool_result:.*|)\s*#\]'
        blob_regex = r'\[#\s*blob\s*(.*?)\s*#\]'
        tool_call_regex = r'\[#\s*tool_call:\s*(.*?)\s*#\]'

        # Check if the raw template contains message prompt sections
        if re.search(regex, full_prompt):
            # Split the full prompt into sections
            sections = re.split(regex, full_prompt)[1:]
            messages = []
            for i in range(0, len(sections), 2):
                role = sections[i].strip()
                content = sections[i + 1].strip()

                # Check if tool result
                if role.startswith('tool_result:'):
                    id = role.split(':')[1].strip()
                    message = ToolResultMessage(content=content.strip(), role='tool', id=id)
                    messages.append(message)
                    continue

                content_parts = []
                tool_calls = []

                # todo can't use images and call tools at the same time right now
                # Check for blobs (images, videos, ...) in the content and replace with Blob objects
                if re.search(blob_regex, content):
                    start = 0
                    for match in re.finditer(blob_regex, content):
                        # Add the text before the blob
                        before_text = content[start:match.start()].strip()
                        if before_text:
                            content_parts.append(before_text)
                        # Add the blob
                        blob_url = match.group(1)
                        content_parts.append(
                            Blob.from_path_or_url(blob_url)
                        )
                        start = match.end()

                    # Add the remaining text after the last blob
                    content_parts.append(content[start:].strip())

                # Check for tool calls in the content
                elif re.search(tool_call_regex, content):
                    # Split the content into tool call sections
                    tool_call_sections = re.split(tool_call_regex, content)
                    # The content before the first tool call is the message content
                    content_parts.append(tool_call_sections[0].strip())
                    tool_call_sections = tool_call_sections[1:]

                    for j in range(0, len(tool_call_sections), 2):
                        tool_call_info = tool_call_sections[j]
                        tool_call_content = tool_call_sections[j + 1].strip()

                        tool_call_pattern = r"^(.+?)\s*\((.+)\)$"
                        match = re.search(tool_call_pattern, tool_call_info)
                        if match:
                            tool_name = match.group(1)
                            tool_call_id = match.group(2)
                        else:
                            tool_name = tool_call_info
                            tool_call_id = None

                        tool_calls.append(ToolCallMessage(
                            id=(tool_call_id or tool_name).strip(),
                            name=tool_name.strip(),
                            payload=tool_call_content.strip())
                        )
                else:
                    # If there are no tool calls or images, the entire content is the message content
                    content_parts.append(content.strip())

                # Create the message object
                content = content_parts[0] if len(content_parts) == 1 else (content_parts if len(content_parts) > 0 else None)
                if role == "assistant":
                    message = AssistantMessage(content=content, tool_calls=tool_calls if tool_calls else None)
                elif role == "system":
                    message = SystemMessage(content=content)
                else:
                    message = UserMessage(content=content)

                messages.append(message)
        else:
            # If the raw template doesn't contain any sections, treat whole file as 'user' message
            messages = [UserMessage(content=full_prompt)]

        return [m for m in messages if m]

    @staticmethod
    def load_template(path: str, llm_name: str) -> Template:
        default_path = f"{path}.prompt"
        llm_specific_path = f"{path}.{llm_name}.prompt"
        filename = llm_specific_path if os.path.exists(llm_specific_path) else default_path
        if not os.path.exists(filename):
            raise FileNotFoundError(f"No prompt file found at {filename}")
        with open(filename, 'r') as f:
            return Template(f.read())



from typing import Any, Dict, List, Union, Callable

TYPE_MAP = {
    int: 'integer',
    float: 'number',
    bool: 'boolean',
    None: None
}


def is_schema(object_: Any) -> bool:
    return 'type' in object_


def handle_schema(json: Dict[str, Any], schema: Dict[str, Any]) -> None:
    schema.update(json)
    if schema['type'] == 'object':
        schema.pop('properties', None)
        parse(json.get('properties', {}), schema)
    if schema['type'] == 'array':
        schema.pop('items', None)
        schema['items'] = {}
        parse(json.get('items', {}), schema['items'])


def handle_array(arr: List[Any], schema: Dict[str, Any]) -> None:
    schema['type'] = 'array'
    props = schema['items'] = {}
    parse(arr[0], props)


def handle_object(json: Dict[str, Any], schema: Dict[str, Any]) -> None:
    if is_schema(json):
        handle_schema(json, schema)
        return
    schema['type'] = 'object'
    schema['required'] = list(json.keys())
    schema['properties'] = {k: {} for k in json.keys()}
    for key, item in json.items():
        parse(item, schema['properties'][key])


def parse(json: Union[Dict[str, Any], List[Any]], schema: Dict[str, Any]) -> None:
    if isinstance(json, list):
        handle_array(json, schema)
    elif isinstance(json, dict):
        handle_object(json, schema)
    elif isinstance(json, str):
        schema['type'] = 'string'
        schema['description'] = json
    else:
        schema['type'] = TYPE_MAP.get(type(json), 'string')
        schema['example'] = json


def expand_simplified_json_schema(data: Dict[str, Any]) -> Dict[str, Any]:
    json_schema = {}
    parse(data, json_schema)
    return json_schema


def sanitize_schema(schema, sanitize_key: Callable, key_mapping: dict):
    if isinstance(schema, dict):
        modified_schema = {}
        for key, value in schema.items():
            if key == 'required':
                # Process 'required' list separately
                modified_schema[key] = [sanitize_key(item) for item in value]
            else:
                sanitized_key = sanitize_key(key)
                if sanitized_key != key:
                    key_mapping[sanitized_key] = key
                # Recursively process nested dictionaries
                if isinstance(value, dict):
                    modified_schema[sanitized_key] = sanitize_schema(value, sanitize_key, key_mapping)
                else:
                    modified_schema[sanitized_key] = value
        return modified_schema
    else:
        # Return non-dictionary values as is
        return schema


def desanitize_json(json, key_mapping: dict):
    if isinstance(json, dict):
        original_schema = {}
        for key, value in json.items():
            original_key = key_mapping.get(key, key)
            # Recursively process nested dictionaries
            if isinstance(value, dict):
                original_schema[original_key] = desanitize_json(value, key_mapping)
            else:
                original_schema[original_key] = value
        return original_schema
    else:
        # Return non-dictionary values as is
        return json



import logging
import os
import re
from io import BytesIO

from typing_extensions import override

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.question_answering.question_answering import QuestionAnswering, QAResult
from bpm_ai_core.util.image import blob_as_images
from bpm_ai_core.util.linguistics import stopwords

try:
    from azure.ai.documentintelligence.aio import DocumentIntelligenceClient as AsyncDocumentIntelligenceClient
    from azure.ai.documentintelligence.models import DocumentAnalysisFeature
    from azure.core.credentials import AzureKeyCredential

    has_azure_doc = True
except ImportError:
    has_azure_doc = False

azure_logger = logging.getLogger('azure')
azure_logger.setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

IMAGE_FORMATS = ["png", "jpeg", "tiff"]


class AzureDocVQA(QuestionAnswering):
    def __init__(self, endpoint: str = None):
        if not has_azure_doc:
            raise ImportError('azure-ai-documentintelligence is not installed')
        self.endpoint = endpoint or os.environ.get("AZURE_DOCUMENT_INTELLIGENCE_ENDPOINT")

    @override
    async def _do_answer(
            self,
            context_str_or_blob: str | Blob,
            question: str
    ) -> QAResult:
        if isinstance(context_str_or_blob, str) or not (context_str_or_blob.is_image() or context_str_or_blob.is_pdf()):
            raise ValueError("Blob must be a PDF or an image")
        if context_str_or_blob.is_pdf():
            bytes_io = await context_str_or_blob.as_bytes_io()
        else:
            bytes_io = BytesIO((await blob_as_images(context_str_or_blob, accept_formats=IMAGE_FORMATS, return_bytes=True))[0])

        question = self.to_camel_case(question)
        logger.info(f"Modified query: '{question}'")

        async with AsyncDocumentIntelligenceClient(
            self.endpoint,
            AzureKeyCredential(
                os.environ.get("AZURE_DOCUMENT_INTELLIGENCE_KEY")
            )
        ) as client:
            document = await client.begin_analyze_document(
                model_id="prebuilt-layout",
                analyze_request=bytes_io,
                content_type="application/octet-stream",
                features=[DocumentAnalysisFeature.QUERY_FIELDS],
                query_fields=[question]
            )

            # Wait for the extraction to complete asynchronously
            result = await document.result()
            prediction = result['documents'][0]['fields'][question]

        return QAResult(
            answer=prediction['content'],
            score=prediction['confidence'],
            start_index=None,
            end_index=None,
        )

    @staticmethod
    def to_camel_case(text: str) -> str:
        text = text.lower()
        text = re.sub(r'[^a-zA-Z0-9\s]', '', text)
        words = [word for word in text.split() if word not in stopwords]
        camel_case_words = [words[0]] + [word.capitalize() for word in words[1:]]
        return ''.join(camel_case_words)



stopwords = ['au', 'aux', 'avec', 'ce', 'ces', 'dans', 'de', 'des', 'du', 'elle', 'en', 'et', 'eux', 'il', 'ils',
             'je', 'la', 'le', 'les', 'leur', 'lui', 'ma', 'mais', 'me', 'même', 'mes', 'moi', 'mon', 'ne', 'nos',
             'notre', 'nous', 'on', 'ou', 'par', 'pas', 'pour', 'qu', 'que', 'qui', 'sa', 'se', 'ses', 'son',
             'sur', 'ta', 'te', 'tes', 'toi', 'ton', 'tu', 'un', 'une', 'vos', 'votre', 'vous', 'c', 'd', 'j', 'l',
             'à', 'm', 'n', 's', 't', 'y', 'été', 'étée', 'étées', 'étés', 'étant', 'étante', 'étants', 'étantes',
             'suis', 'es', 'est', 'sommes', 'êtes', 'sont', 'serai', 'seras', 'sera', 'serons', 'serez', 'seront',
             'serais', 'serait', 'serions', 'seriez', 'seraient', 'étais', 'était', 'étions', 'étiez', 'étaient',
             'fus', 'fut', 'fûmes', 'fûtes', 'furent', 'sois', 'soit', 'soyons', 'soyez', 'soient', 'fusse',
             'fusses', 'fût', 'fussions', 'fussiez', 'fussent', 'ayant', 'ayante', 'ayantes', 'ayants', 'eu',
             'eue', 'eues', 'eus', 'ai', 'as', 'avons', 'avez', 'ont', 'aurai', 'auras', 'aura', 'aurons', 'aurez',
             'auront', 'aurais', 'aurait', 'aurions', 'auriez', 'auraient', 'avais', 'avait', 'avions', 'aviez',
             'avaient', 'eut', 'eûmes', 'eûtes', 'eurent', 'aie', 'aies', 'ait', 'ayons', 'ayez', 'aient', 'eusse',
             'eusses', 'eût', 'eussions', 'eussiez', 'eussent', 'og', 'i', 'jeg', 'det', 'at', 'en', 'den', 'til',
             'er', 'som', 'på', 'de', 'med', 'han', 'af', 'for', 'ikke', 'der', 'var', 'mig', 'sig', 'men', 'et',
             'har', 'om', 'vi', 'min', 'havde', 'ham', 'hun', 'nu', 'over', 'da', 'fra', 'du', 'ud', 'sin', 'dem',
             'os', 'op', 'man', 'hans', 'hvor', 'eller', 'hvad', 'skal', 'selv', 'her', 'alle', 'vil', 'blev',
             'kunne', 'ind', 'når', 'være', 'dog', 'noget', 'ville', 'jo', 'deres', 'efter', 'ned', 'skulle',
             'denne', 'end', 'dette', 'mit', 'også', 'under', 'have', 'dig', 'anden', 'hende', 'mine', 'alt',
             'meget', 'sit', 'sine', 'vor', 'mod', 'disse', 'hvis', 'din', 'nogle', 'hos', 'blive', 'mange', 'ad',
             'bliver', 'hendes', 'været', 'thi', 'jer', 'sådan', 'olla', 'olen', 'olet', 'on', 'olemme', 'olette',
             'ovat', 'ole', 'oli', 'olisi', 'olisit', 'olisin', 'olisimme', 'olisitte', 'olisivat', 'olit', 'olin',
             'olimme', 'olitte', 'olivat', 'ollut', 'olleet', 'en', 'et', 'ei', 'emme', 'ette', 'eivät', 'minä',
             'minun', 'minut', 'minua', 'minussa', 'minusta', 'minuun', 'minulla', 'minulta', 'minulle', 'sinä',
             'sinun', 'sinut', 'sinua', 'sinussa', 'sinusta', 'sinuun', 'sinulla', 'sinulta', 'sinulle', 'hän',
             'hänen', 'hänet', 'häntä', 'hänessä', 'hänestä', 'häneen', 'hänellä', 'häneltä', 'hänelle', 'me',
             'meidän', 'meidät', 'meitä', 'meissä', 'meistä', 'meihin', 'meillä', 'meiltä', 'meille', 'te',
             'teidän', 'teidät', 'teitä', 'teissä', 'teistä', 'teihin', 'teillä', 'teiltä', 'teille', 'he',
             'heidän', 'heidät', 'heitä', 'heissä', 'heistä', 'heihin', 'heillä', 'heiltä', 'heille', 'tämä',
             'tämän', 'tätä', 'tässä', 'tästä', 'tähän', 'tallä', 'tältä', 'tälle', 'tänä', 'täksi', 'tuo', 'tuon',
             'tuotä', 'tuossa', 'tuosta', 'tuohon', 'tuolla', 'tuolta', 'tuolle', 'tuona', 'tuoksi', 'se', 'sen',
             'sitä', 'siinä', 'siitä', 'siihen', 'sillä', 'siltä', 'sille', 'sinä', 'siksi', 'nämä', 'näiden',
             'näitä', 'näissä', 'näistä', 'näihin', 'näillä', 'näiltä', 'näille', 'näinä', 'näiksi', 'nuo',
             'noiden', 'noita', 'noissa', 'noista', 'noihin', 'noilla', 'noilta', 'noille', 'noina', 'noiksi',
             'ne', 'niiden', 'niitä', 'niissä', 'niistä', 'niihin', 'niillä', 'niiltä', 'niille', 'niinä',
             'niiksi', 'kuka', 'kenen', 'kenet', 'ketä', 'kenessä', 'kenestä', 'keneen', 'kenellä', 'keneltä',
             'kenelle', 'kenenä', 'keneksi', 'ketkä', 'keiden', 'ketkä', 'keitä', 'keissä', 'keistä', 'keihin',
             'keillä', 'keiltä', 'keille', 'keinä', 'keiksi', 'mikä', 'minkä', 'minkä', 'mitä', 'missä', 'mistä',
             'mihin', 'millä', 'miltä', 'mille', 'minä', 'miksi', 'mitkä', 'joka', 'jonka', 'jota', 'jossa',
             'josta', 'johon', 'jolla', 'jolta', 'jolle', 'jona', 'joksi', 'jotka', 'joiden', 'joita', 'joissa',
             'joista', 'joihin', 'joilla', 'joilta', 'joille', 'joina', 'joiksi', 'että', 'ja', 'jos', 'koska',
             'kuin', 'mutta', 'niin', 'sekä', 'sillä', 'tai', 'vaan', 'vai', 'vaikka', 'kanssa', 'mukaan', 'noin',
             'poikki', 'yli', 'kun', 'niin', 'nyt', 'itse', 'a', 'à', 'ao', 'aos', 'aquela', 'aquelas', 'aquele',
             'aqueles', 'aquilo', 'as', 'às', 'até', 'com', 'como', 'da', 'das', 'de', 'dela', 'delas', 'dele',
             'deles', 'depois', 'do', 'dos', 'e', 'é', 'ela', 'elas', 'ele', 'eles', 'em', 'entre', 'era', 'eram',
             'éramos', 'essa', 'essas', 'esse', 'esses', 'esta', 'está', 'estamos', 'estão', 'estar', 'estas',
             'estava', 'estavam', 'estávamos', 'este', 'esteja', 'estejam', 'estejamos', 'estes', 'esteve',
             'estive', 'estivemos', 'estiver', 'estivera', 'estiveram', 'estivéramos', 'estiverem', 'estivermos',
             'estivesse', 'estivessem', 'estivéssemos', 'estou', 'eu', 'foi', 'fomos', 'for', 'fora', 'foram',
             'fôramos', 'forem', 'formos', 'fosse', 'fossem', 'fôssemos', 'fui', 'há', 'haja', 'hajam', 'hajamos',
             'hão', 'havemos', 'haver', 'hei', 'houve', 'houvemos', 'houver', 'houvera', 'houverá', 'houveram',
             'houvéramos', 'houverão', 'houverei', 'houverem', 'houveremos', 'houveria', 'houveriam',
             'houveríamos', 'houvermos', 'houvesse', 'houvessem', 'houvéssemos', 'isso', 'isto', 'já', 'lhe',
             'lhes', 'mais', 'mas', 'me', 'mesmo', 'meu', 'meus', 'minha', 'minhas', 'muito', 'na', 'não', 'nas',
             'nem', 'no', 'nos', 'nós', 'nossa', 'nossas', 'nosso', 'nossos', 'num', 'numa', 'o', 'os', 'ou',
             'para', 'pela', 'pelas', 'pelo', 'pelos', 'por', 'qual', 'quando', 'que', 'quem', 'são', 'se', 'seja',
             'sejam', 'sejamos', 'sem', 'ser', 'será', 'serão', 'serei', 'seremos', 'seria', 'seriam', 'seríamos',
             'seu', 'seus', 'só', 'somos', 'sou', 'sua', 'suas', 'também', 'te', 'tem', 'tém', 'temos', 'tenha',
             'tenham', 'tenhamos', 'tenho', 'terá', 'terão', 'terei', 'teremos', 'teria', 'teriam', 'teríamos',
             'teu', 'teus', 'teve', 'tinha', 'tinham', 'tínhamos', 'tive', 'tivemos', 'tiver', 'tivera', 'tiveram',
             'tivéramos', 'tiverem', 'tivermos', 'tivesse', 'tivessem', 'tivéssemos', 'tu', 'tua', 'tuas', 'um',
             'uma', 'você', 'vocês', 'vos', 'og', 'i', 'jeg', 'det', 'at', 'en', 'et', 'den', 'til', 'er', 'som',
             'på', 'de', 'med', 'han', 'av', 'ikke', 'ikkje', 'der', 'så', 'var', 'meg', 'seg', 'men', 'ett',
             'har', 'om', 'vi', 'min', 'mitt', 'ha', 'hadde', 'hun', 'nå', 'over', 'da', 'ved', 'fra', 'du', 'ut',
             'sin', 'dem', 'oss', 'opp', 'man', 'kan', 'hans', 'hvor', 'eller', 'hva', 'skal', 'selv', 'sjøl',
             'her', 'alle', 'vil', 'bli', 'ble', 'blei', 'blitt', 'kunne', 'inn', 'når', 'være', 'kom', 'noen',
             'noe', 'ville', 'dere', 'som', 'deres', 'kun', 'ja', 'etter', 'ned', 'skulle', 'denne', 'for', 'deg',
             'si', 'sine', 'sitt', 'mot', 'å', 'meget', 'hvorfor', 'dette', 'disse', 'uten', 'hvordan', 'ingen',
             'din', 'ditt', 'blir', 'samme', 'hvilken', 'hvilke', 'sånn', 'inni', 'mellom', 'vår', 'hver', 'hvem',
             'vors', 'hvis', 'både', 'bare', 'enn', 'fordi', 'før', 'mange', 'også', 'slik', 'vært', 'være', 'båe',
             'begge', 'siden', 'dykk', 'dykkar', 'dei', 'deira', 'deires', 'deim', 'di', 'då', 'eg', 'ein', 'eit',
             'eitt', 'elles', 'honom', 'hjå', 'ho', 'hoe', 'henne', 'hennar', 'hennes', 'hoss', 'hossen', 'ikkje',
             'ingi', 'inkje', 'korleis', 'korso', 'kva', 'kvar', 'kvarhelst', 'kven', 'kvi', 'kvifor', 'me',
             'medan', 'mi', 'mine', 'mykje', 'no', 'nokon', 'noka', 'nokor', 'noko', 'nokre', 'si', 'sia', 'sidan',
             'so', 'somt', 'somme', 'um', 'upp', 'vere', 'vore', 'verte', 'vort', 'varte', 'vart', 'i', 'me', 'my',
             'myself', 'we', 'our', 'ours', 'ourselves', 'you', "you're", "you've", "you'll", "you'd", 'your',
             'yours', 'yourself', 'yourselves', 'he', 'him', 'his', 'himself', 'she', "she's", 'her', 'hers',
             'herself', 'it', "it's", 'its', 'itself', 'they', 'them', 'their', 'theirs', 'themselves', 'what',
             'which', 'who', 'whom', 'this', 'that', "that'll", 'these', 'those', 'am', 'is', 'are', 'was', 'were',
             'be', 'been', 'being', 'have', 'has', 'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the',
             'and', 'but', 'if', 'or', 'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for', 'with', 'about',
             'against', 'between', 'into', 'through', 'during', 'before', 'after', 'above', 'below', 'to', 'from',
             'up', 'down', 'in', 'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once', 'here',
             'there', 'when', 'where', 'why', 'how', 'all', 'any', 'both', 'each', 'few', 'more', 'most', 'other',
             'some', 'such', 'no', 'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too', 'very', 's', 't',
             'can', 'will', 'just', 'don', "don't", 'should', "should've", 'now', 'd', 'll', 'm', 'o', 're', 've',
             'y', 'ain', 'aren', "aren't", 'couldn', "couldn't", 'didn', "didn't", 'doesn', "doesn't", 'hadn',
             "hadn't", 'hasn', "hasn't", 'haven', "haven't", 'isn', "isn't", 'ma', 'mightn', "mightn't", 'mustn',
             "mustn't", 'needn', "needn't", 'shan', "shan't", 'shouldn', "shouldn't", 'wasn', "wasn't", 'weren',
             "weren't", 'won', "won't", 'wouldn', "wouldn't", 'ad', 'al', 'allo', 'ai', 'agli', 'all', 'agl',
             'alla', 'alle', 'con', 'col', 'coi', 'da', 'dal', 'dallo', 'dai', 'dagli', 'dall', 'dagl', 'dalla',
             'dalle', 'di', 'del', 'dello', 'dei', 'degli', 'dell', 'degl', 'della', 'delle', 'in', 'nel', 'nello',
             'nei', 'negli', 'nell', 'negl', 'nella', 'nelle', 'su', 'sul', 'sullo', 'sui', 'sugli', 'sull',
             'sugl', 'sulla', 'sulle', 'per', 'tra', 'contro', 'io', 'tu', 'lui', 'lei', 'noi', 'voi', 'loro',
             'mio', 'mia', 'miei', 'mie', 'tuo', 'tua', 'tuoi', 'tue', 'suo', 'sua', 'suoi', 'sue', 'nostro',
             'nostra', 'nostri', 'nostre', 'vostro', 'vostra', 'vostri', 'vostre', 'mi', 'ti', 'ci', 'vi', 'lo',
             'la', 'li', 'le', 'gli', 'ne', 'il', 'un', 'uno', 'una', 'ma', 'ed', 'se', 'perché', 'anche', 'come',
             'dov', 'dove', 'che', 'chi', 'cui', 'non', 'più', 'quale', 'quanto', 'quanti', 'quanta', 'quante',
             'quello', 'quelli', 'quella', 'quelle', 'questo', 'questi', 'questa', 'queste', 'si', 'tutto',
             'tutti', 'a', 'c', 'e', 'i', 'l', 'o', 'ho', 'hai', 'ha', 'abbiamo', 'avete', 'hanno', 'abbia',
             'abbiate', 'abbiano', 'avrò', 'avrai', 'avrà', 'avremo', 'avrete', 'avranno', 'avrei', 'avresti',
             'avrebbe', 'avremmo', 'avreste', 'avrebbero', 'avevo', 'avevi', 'aveva', 'avevamo', 'avevate',
             'avevano', 'ebbi', 'avesti', 'ebbe', 'avemmo', 'aveste', 'ebbero', 'avessi', 'avesse', 'avessimo',
             'avessero', 'avendo', 'avuto', 'avuta', 'avuti', 'avute', 'sono', 'sei', 'è', 'siamo', 'siete', 'sia',
             'siate', 'siano', 'sarò', 'sarai', 'sarà', 'saremo', 'sarete', 'saranno', 'sarei', 'saresti',
             'sarebbe', 'saremmo', 'sareste', 'sarebbero', 'ero', 'eri', 'era', 'eravamo', 'eravate', 'erano',
             'fui', 'fosti', 'fu', 'fummo', 'foste', 'furono', 'fossi', 'fosse', 'fossimo', 'fossero', 'essendo',
             'faccio', 'fai', 'facciamo', 'fanno', 'faccia', 'facciate', 'facciano', 'farò', 'farai', 'farà',
             'faremo', 'farete', 'faranno', 'farei', 'faresti', 'farebbe', 'faremmo', 'fareste', 'farebbero',
             'facevo', 'facevi', 'faceva', 'facevamo', 'facevate', 'facevano', 'feci', 'facesti', 'fece',
             'facemmo', 'faceste', 'fecero', 'facessi', 'facesse', 'facessimo', 'facessero', 'facendo', 'sto',
             'stai', 'sta', 'stiamo', 'stanno', 'stia', 'stiate', 'stiano', 'starò', 'starai', 'starà', 'staremo',
             'starete', 'staranno', 'starei', 'staresti', 'starebbe', 'staremmo', 'stareste', 'starebbero',
             'stavo', 'stavi', 'stava', 'stavamo', 'stavate', 'stavano', 'stetti', 'stesti', 'stette', 'stemmo',
             'steste', 'stettero', 'stessi', 'stesse', 'stessimo', 'stessero', 'stando', 'acaba', 'ama', 'aslında',
             'az', 'bazı', 'belki', 'biri', 'birkaç', 'birşey', 'biz', 'bu', 'çok', 'çünkü', 'da', 'daha', 'de',
             'defa', 'diye', 'eğer', 'en', 'gibi', 'hem', 'hep', 'hepsi', 'her', 'hiç', 'için', 'ile', 'ise',
             'kez', 'ki', 'kim', 'mı', 'mu', 'mü', 'nasıl', 'ne', 'neden', 'nerde', 'nerede', 'nereye', 'niçin',
             'niye', 'o', 'sanki', 'şey', 'siz', 'şu', 'tüm', 've', 'veya', 'ya', 'yani', 'aber', 'alle', 'allem',
             'allen', 'aller', 'alles', 'als', 'also', 'am', 'an', 'ander', 'andere', 'anderem', 'anderen',
             'anderer', 'anderes', 'anderm', 'andern', 'anderr', 'anders', 'auch', 'auf', 'aus', 'bei', 'bin',
             'bis', 'bist', 'da', 'damit', 'dann', 'der', 'den', 'des', 'dem', 'die', 'das', 'dass', 'daß',
             'derselbe', 'derselben', 'denselben', 'desselben', 'demselben', 'dieselbe', 'dieselben', 'dasselbe',
             'dazu', 'dein', 'deine', 'deinem', 'deinen', 'deiner', 'deines', 'denn', 'derer', 'dessen', 'dich',
             'dir', 'du', 'dies', 'diese', 'diesem', 'diesen', 'dieser', 'dieses', 'doch', 'dort', 'durch', 'ein',
             'eine', 'einem', 'einen', 'einer', 'eines', 'einig', 'einige', 'einigem', 'einigen', 'einiger',
             'einiges', 'einmal', 'er', 'ihn', 'ihm', 'es', 'etwas', 'euer', 'eure', 'eurem', 'euren', 'eurer',
             'eures', 'für', 'gegen', 'gewesen', 'hab', 'habe', 'haben', 'hat', 'hatte', 'hatten', 'hier', 'hin',
             'hinter', 'ich', 'mich', 'mir', 'ihr', 'ihre', 'ihrem', 'ihren', 'ihrer', 'ihres', 'euch', 'im', 'in',
             'indem', 'ins', 'ist', 'jede', 'jedem', 'jeden', 'jeder', 'jedes', 'jene', 'jenem', 'jenen', 'jener',
             'jenes', 'jetzt', 'kann', 'kein', 'keine', 'keinem', 'keinen', 'keiner', 'keines', 'können', 'könnte',
             'machen', 'man', 'manche', 'manchem', 'manchen', 'mancher', 'manches', 'mein', 'meine', 'meinem',
             'meinen', 'meiner', 'meines', 'mit', 'muss', 'musste', 'nach', 'nicht', 'nichts', 'noch', 'nun',
             'nur', 'ob', 'oder', 'ohne', 'sehr', 'sein', 'seine', 'seinem', 'seinen', 'seiner', 'seines',
             'selbst', 'sich', 'sie', 'ihnen', 'sind', 'so', 'solche', 'solchem', 'solchen', 'solcher', 'solches',
             'soll', 'sollte', 'sondern', 'sonst', 'über', 'um', 'und', 'uns', 'unsere', 'unserem', 'unseren',
             'unser', 'unseres', 'unter', 'viel', 'vom', 'von', 'vor', 'während', 'war', 'waren', 'warst', 'was',
             'weg', 'weil', 'weiter', 'welche', 'welchem', 'welchen', 'welcher', 'welches', 'wenn', 'werde',
             'werden', 'wie', 'wieder', 'will', 'wir', 'wird', 'wirst', 'wo', 'wollen', 'wollte', 'würde',
             'würden', 'zu', 'zum', 'zur', 'zwar', 'zwischen', 'de', 'en', 'van', 'ik', 'te', 'dat', 'die', 'in',
             'een', 'hij', 'het', 'niet', 'zijn', 'is', 'was', 'op', 'aan', 'met', 'als', 'voor', 'had', 'er',
             'maar', 'om', 'hem', 'dan', 'zou', 'of', 'wat', 'mijn', 'men', 'dit', 'zo', 'door', 'over', 'ze',
             'zich', 'bij', 'ook', 'tot', 'je', 'mij', 'uit', 'der', 'daar', 'haar', 'naar', 'heb', 'hoe', 'heeft',
             'hebben', 'deze', 'u', 'want', 'nog', 'zal', 'me', 'zij', 'nu', 'ge', 'geen', 'omdat', 'iets',
             'worden', 'toch', 'al', 'waren', 'veel', 'meer', 'doen', 'toen', 'moet', 'ben', 'zonder', 'kan',
             'hun', 'dus', 'alles', 'onder', 'ja', 'eens', 'hier', 'wie', 'werd', 'altijd', 'doch', 'wordt',
             'wezen', 'kunnen', 'ons', 'zelf', 'tegen', 'na', 'reeds', 'wil', 'kon', 'niets', 'uw', 'iemand',
             'geweest', 'andere', 'och', 'det', 'att', 'i', 'en', 'jag', 'hon', 'som', 'han', 'på', 'den', 'med',
             'var', 'sig', 'för', 'så', 'till', 'är', 'men', 'ett', 'om', 'hade', 'de', 'av', 'icke', 'mig', 'du',
             'henne', 'då', 'sin', 'nu', 'har', 'inte', 'hans', 'honom', 'skulle', 'hennes', 'där', 'min', 'man',
             'ej', 'vid', 'kunde', 'något', 'från', 'ut', 'när', 'efter', 'upp', 'vi', 'dem', 'vara', 'vad',
             'över', 'än', 'dig', 'kan', 'sina', 'här', 'ha', 'mot', 'alla', 'under', 'någon', 'eller', 'allt',
             'mycket', 'sedan', 'ju', 'denna', 'själv', 'detta', 'åt', 'utan', 'varit', 'hur', 'ingen', 'mitt',
             'ni', 'bli', 'blev', 'oss', 'din', 'dessa', 'några', 'deras', 'blir', 'mina', 'samma', 'vilken', 'er',
             'sådan', 'vår', 'blivit', 'dess', 'inom', 'mellan', 'sådant', 'varför', 'varje', 'vilka', 'ditt',
             'vem', 'vilket', 'sitta', 'sådana', 'vart', 'dina', 'vars', 'vårt', 'våra', 'ert', 'era', 'vilkas',
             'de', 'la', 'que', 'el', 'en', 'y', 'a', 'los', 'del', 'se', 'las', 'por', 'un', 'para', 'con', 'no',
             'una', 'su', 'al', 'lo', 'como', 'más', 'pero', 'sus', 'le', 'ya', 'o', 'este', 'sí', 'porque',
             'esta', 'entre', 'cuando', 'muy', 'sin', 'sobre', 'también', 'me', 'hasta', 'hay', 'donde', 'quien',
             'desde', 'todo', 'nos', 'durante', 'todos', 'uno', 'les', 'ni', 'contra', 'otros', 'ese', 'eso',
             'ante', 'ellos', 'e', 'esto', 'mí', 'antes', 'algunos', 'qué', 'unos', 'yo', 'otro', 'otras', 'otra',
             'él', 'tanto', 'esa', 'estos', 'mucho', 'quienes', 'nada', 'muchos', 'cual', 'poco', 'ella', 'estar',
             'estas', 'algunas', 'algo', 'nosotros', 'mi', 'mis', 'tú', 'te', 'ti', 'tu', 'tus', 'ellas',
             'nosotras', 'vosotros', 'vosotras', 'os', 'mío', 'mía', 'míos', 'mías', 'tuyo', 'tuya', 'tuyos',
             'tuyas', 'suyo', 'suya', 'suyos', 'suyas', 'nuestro', 'nuestra', 'nuestros', 'nuestras', 'vuestro',
             'vuestra', 'vuestros', 'vuestras', 'esos', 'esas', 'estoy', 'estás', 'está', 'estamos', 'estáis',
             'están', 'esté', 'estés', 'estemos', 'estéis', 'estén', 'estaré', 'estarás', 'estará', 'estaremos',
             'estaréis', 'estarán', 'estaría', 'estarías', 'estaríamos', 'estaríais', 'estarían', 'estaba',
             'estabas', 'estábamos', 'estabais', 'estaban', 'estuve', 'estuviste', 'estuvo', 'estuvimos',
             'estuvisteis', 'estuvieron', 'estuviera', 'estuvieras', 'estuviéramos', 'estuvierais', 'estuvieran',
             'estuviese', 'estuvieses', 'estuviésemos', 'estuvieseis', 'estuviesen', 'estando', 'estado', 'estada',
             'estados', 'estadas', 'estad', 'he', 'has', 'ha', 'hemos', 'habéis', 'han', 'haya', 'hayas',
             'hayamos', 'hayáis', 'hayan', 'habré', 'habrás', 'habrá', 'habremos', 'habréis', 'habrán', 'habría',
             'habrías', 'habríamos', 'habríais', 'habrían', 'había', 'habías', 'habíamos', 'habíais', 'habían',
             'hube', 'hubiste', 'hubo', 'hubimos', 'hubisteis', 'hubieron', 'hubiera', 'hubieras', 'hubiéramos',
             'hubierais', 'hubieran', 'hubiese', 'hubieses', 'hubiésemos', 'hubieseis', 'hubiesen', 'habiendo',
             'habido', 'habida', 'habidos', 'habidas', 'soy', 'eres', 'es', 'somos', 'sois', 'son', 'sea', 'seas',
             'seamos', 'seáis', 'sean', 'seré', 'serás', 'será', 'seremos', 'seréis', 'serán', 'sería', 'serías',
             'seríamos', 'seríais', 'serían', 'era', 'eras', 'éramos', 'erais', 'eran', 'fui', 'fuiste', 'fue',
             'fuimos', 'fuisteis', 'fueron', 'fuera', 'fueras', 'fuéramos', 'fuerais', 'fueran', 'fuese', 'fueses',
             'fuésemos', 'fueseis', 'fuesen', 'sintiendo', 'sentido', 'sentida', 'sentidos', 'sentidas', 'siente',
             'sentid', 'tengo', 'tienes', 'tiene', 'tenemos', 'tenéis', 'tienen', 'tenga', 'tengas', 'tengamos',
             'tengáis', 'tengan', 'tendré', 'tendrás', 'tendrá', 'tendremos', 'tendréis', 'tendrán', 'tendría',
             'tendrías', 'tendríamos', 'tendríais', 'tendrían', 'tenía', 'tenías', 'teníamos', 'teníais', 'tenían',
             'tuve', 'tuviste', 'tuvo', 'tuvimos', 'tuvisteis', 'tuvieron', 'tuviera', 'tuvieras', 'tuviéramos',
             'tuvierais', 'tuvieran', 'tuviese', 'tuvieses', 'tuviésemos', 'tuvieseis', 'tuviesen', 'teniendo',
             'tenido', 'tenida', 'tenidos', 'tenidas', 'tened']


special_char_map = {'ä': 'ae', 'Ä': 'Ae', 'ö': 'oe', 'Ö': 'Oe', 'ü': 'ue', 'Ü': 'Ue', 'ß': 'ss', 'æ': 'ae', 'Æ': 'Ae',
                    'ø': 'oe', 'Ø': 'Oe', 'ë': 'e', 'Ë': 'E', 'ï': 'i', 'Ï': 'I', 'à': 'a', 'À': 'A', 'â': 'a',
                    'Â': 'A', 'ç': 'c', 'Ç': 'C', 'é': 'e', 'É': 'E', 'è': 'e', 'È': 'E', 'ê': 'e', 'Ê': 'E', 'î': 'i',
                    'Î': 'I', 'ô': 'o', 'Ô': 'O', 'œ': 'oe', 'Œ': 'Oe', 'ù': 'u', 'Ù': 'U', 'û': 'u', 'Û': 'U',
                    'ÿ': 'y', 'Ÿ': 'Y', 'ì': 'i', 'Ì': 'I', 'ò': 'o', 'Ò': 'O', 'ą': 'a', 'Ą': 'A', 'ć': 'c', 'Ć': 'C',
                    'ę': 'e', 'Ę': 'E', 'ł': 'l', 'Ł': 'L', 'ń': 'n', 'Ń': 'N', 'ó': 'o', 'Ó': 'O', 'ś': 's', 'Ś': 'S',
                    'ź': 'z', 'Ź': 'Z', 'ż': 'z', 'Ż': 'Z', 'á': 'a', 'Á': 'A', 'ã': 'a', 'Ã': 'A', 'í': 'i', 'Í': 'I',
                    'õ': 'o', 'Õ': 'O', 'ú': 'u', 'Ú': 'U', 'ñ': 'n', 'Ñ': 'N', 'å': 'a', 'Å': 'A', 'ґ': 'g', 'Ґ': 'G',
                    'є': 'ye', 'Є': 'Ye', 'і': 'i', 'І': 'I', 'ї': 'yi', 'Ї': 'Yi'}


def replace_diacritics(text: str):
    return ''.join([special_char_map.get(char, char) for char in text])



import httpx

DEFAULT_MODEL = "claude-3-opus-20240229"
DEFAULT_TEMPERATURE = 0.0
DEFAULT_MAX_RETRIES = 8

default_client_kwargs = {
    "http_client": httpx.AsyncClient(
        limits=httpx.Limits(max_connections=1000, max_keepalive_connections=100)
    ),
    "max_retries": 0  # we use own retry logic
}


from abc import abstractmethod, ABC
from typing import Any, Type

from tenacity import stop_after_attempt, wait_exponential, retry_if_exception_type, Retrying

from bpm_ai_core.llm.common.message import ChatMessage, AssistantMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.tracing.tracing import Tracing


class LLM(ABC):
    """
    Abstract class for large language models.
    """

    def __init__(
        self,
        model: str,
        temperature: float = 0.0,
        max_retries: int = 8,
        retryable_exceptions: list[Type[BaseException]] = None
    ):
        self.model = model
        self.temperature = temperature
        self.max_retries = max_retries
        self.retryable_exceptions = retryable_exceptions or [Exception]

    async def generate_message(
        self,
        prompt: Prompt | list[ChatMessage],
        output_schema: dict[str, Any] = None,
        tools: list[Tool] = None,
        stop: list[str] = None
    ) -> AssistantMessage:
        if output_schema and tools:
            raise ValueError("Must not pass both an output_schema and tools")

        messages = prompt if isinstance(prompt, list) else prompt.format(llm_name=self.name())

        Tracing.tracers().start_span(self.model, inputs={"messages": messages, "output_schema": output_schema, "tools": tools})

        for attempt in Retrying(
            wait=wait_exponential(multiplier=1.5, min=2, max=60),
            stop=stop_after_attempt(self.max_retries),
            retry=retry_if_exception_type(tuple(self.retryable_exceptions))
        ):
            with attempt:
                completion = await self._generate_message(messages, output_schema, tools, stop, attempt.retry_state.attempt_number)

        Tracing.tracers().end_span(outputs=completion.model_dump())
        return completion

    @abstractmethod
    async def _generate_message(
        self,
        messages: list[ChatMessage],
        output_schema: dict[str, Any] = None,
        tools: list[Tool] = None,
        stop: list[str] = None,
        current_try: int = None
    ) -> AssistantMessage:
        pass

    @abstractmethod
    def supports_images(self) -> bool:
        pass

    @abstractmethod
    def supports_video(self) -> bool:
        pass

    @abstractmethod
    def supports_audio(self) -> bool:
        pass

    @abstractmethod
    def name(self) -> str:
        pass



from abc import ABC, abstractmethod
from typing import Tuple

from pydantic import BaseModel

from bpm_ai_core.tracing.decorators import span


class ClassificationResult(BaseModel):
    max_label: str
    max_score: float
    labels_scores: list[Tuple[str, float]]


class ZeroShotClassifier(ABC):
    """
    Zero Shot Classification Model
    """

    @abstractmethod
    async def _do_classify(
            self,
            text: str,
            classes: list[str],
            hypothesis_template: str | None = None,
            multi_label: bool = False
    ) -> ClassificationResult:
        pass

    @span(name="classifier")
    async def classify(
            self,
            text: str,
            classes: list[str],
            confidence_threshold: float | None = None,
            hypothesis_template: str | None = None,
            multi_label: bool = False
    ) -> ClassificationResult:
        result = await self._do_classify(
            text=text,
            classes=classes,
            hypothesis_template=hypothesis_template,
            multi_label=multi_label
        )
        if multi_label:
            result.labels_scores = [(l, s) for l, s in result.labels_scores if s > (confidence_threshold or -1)]

        # Only return if the score is above the threshold (if given)
        return result \
            if result.max_score > (confidence_threshold or -1) \
            else None



import io
import logging
from typing import Optional

from typing_extensions import override

from bpm_ai_core.speech_recognition.asr import ASRModel, ASRResult

logger = logging.getLogger(__name__)

try:
    from openai import AsyncOpenAI, OpenAIError
    import httpx

    has_openai = True
    try:
        client = AsyncOpenAI(
            http_client=httpx.AsyncClient(
                limits=httpx.Limits(max_connections=1000, max_keepalive_connections=100)
            ),
        )
    except OpenAIError as e:
        logger.error(e)
except ImportError:
    has_openai = False


class OpenAIWhisperASR(ASRModel):
    """
    `OpenAI` Whisper Automatic Speech Recognition (ASR) API for transcribing audio.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.
    """

    def __init__(
        self,
        whisper_model: str = "whisper-1"
    ):
        if not has_openai:
            raise ImportError('openai is not installed')
        self.whisper_model = whisper_model

    @override
    async def _do_transcribe(self, audio: io.BytesIO, language: Optional[str] = None) -> ASRResult:
        transcript = await client.audio.transcriptions.create(
            model=self.whisper_model,
            file=audio,
            **{"language": language} if language else {}
        )
        return ASRResult(text=transcript.text)



from abc import ABC, abstractmethod
from typing import Tuple

from pydantic import BaseModel


class POSResult(BaseModel):
    tags: list[Tuple[str, str]]


class POSTagger(ABC):
    """
    Part-of-Speech Tagging Model
    """

    @abstractmethod
    async def _do_tag(self, text: str) -> POSResult:
        pass

    async def tag(self, text: str) -> POSResult:
        """
        Returns a list of tuples (token, tag). Example:
        [('I', 'PRON'), ('am', 'AUX'), ('30', 'NUM'), ('years', 'NOUN'), ('old', 'ADJ'), ('.', 'PUNCT')]
        """
        return await self._do_tag(text)



import json
import logging
from typing import Dict, Any, Optional, List

from typing_extensions import deprecated

from bpm_ai_core.llm.anthropic_chat import get_anthropic_client
from bpm_ai_core.llm.anthropic_chat._constants import DEFAULT_MODEL, DEFAULT_TEMPERATURE, \
    DEFAULT_MAX_RETRIES
from bpm_ai_core.llm.anthropic_chat.util import messages_to_anthropic_dicts, json_schema_to_anthropic_tool, \
    tool_calls_to_tool_message
from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import ChatMessage, ToolCallMessage, AssistantMessage, SystemMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.prompt.prompt import Prompt
from bpm_ai_core.tracing.tracing import Tracing
from bpm_ai_core.util.json_schema import expand_simplified_json_schema

logger = logging.getLogger(__name__)

try:
    from anthropic import AsyncAnthropic, RateLimitError, InternalServerError, APIConnectionError
    from anthropic.types import Message, TextBlock
    from anthropic.types.beta.tools import ToolUseBlock, ToolsBetaMessage

    has_anthropic = True
except ImportError:
    has_anthropic = False


class ChatAnthropic(LLM):
    """
    `Anthropic` Chat large language models API.

    To use, you should have the ``anthropic`` python package installed, and the
    environment variable ``ANTHROPIC_API_KEY`` set with your API key.
    """

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        temperature: float = DEFAULT_TEMPERATURE,
        max_retries: int = DEFAULT_MAX_RETRIES,
        client: AsyncAnthropic = None
    ):
        if not has_anthropic:
            raise ImportError('anthropic is not installed')
        if not client:
            client = get_anthropic_client()
        super().__init__(
            model=model,
            temperature=temperature,
            max_retries=max_retries,
            retryable_exceptions=[
                RateLimitError, InternalServerError, APIConnectionError
            ]
        )
        self.client = client

    @classmethod
    def for_anthropic(
        cls,
        model: str = DEFAULT_MODEL,
        temperature: float = DEFAULT_TEMPERATURE,
        max_retries: int = DEFAULT_MAX_RETRIES
    ):
        return cls(
            model=model,
            temperature=temperature,
            max_retries=max_retries
        )

    async def _generate_message(
        self,
        messages: List[ChatMessage],
        output_schema: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool]] = None,
        stop: list[str] = None,
        current_try: int = None
    ) -> AssistantMessage:
        if output_schema:
            tools = [self._output_schema_to_tool(output_schema)]
            message = await self._run_tool_completion(messages, tools, current_try)
            if message.has_tool_calls():
                return AssistantMessage(content=message.tool_calls[0].payload)
            else:
                return AssistantMessage(content=None)
        elif tools:
            return await self._run_tool_completion(messages, tools, current_try)
        else:
            completion = await self._run_completion(messages, stop, current_try)
            return AssistantMessage(content=completion.content[0].text.strip())

    async def _run_completion(self, messages: List[ChatMessage], stop: list[str] = None, current_try: int = None) -> Message:
        messages = await messages_to_anthropic_dicts(messages)
        Tracing.tracers().start_llm_trace(self, messages, current_try, None)
        completion = await self.client.messages.create(
            max_tokens=4096,
            model=self.model,
            temperature=self.temperature,
            system=messages.pop(0)["content"] if (messages and messages[0]["role"] == "system") else "",
            messages=messages,
            stop_sequences=stop
        )
        Tracing.tracers().end_llm_trace(completion)
        return completion

    async def _run_tool_completion(self, messages: list[ChatMessage], tools: list[Tool] = None, current_try: int = None) -> AssistantMessage:
        sanitized_key_mappings = {t.name: {} for t in tools}
        anthropic_tools = [json_schema_to_anthropic_tool(t.name, t.description, t.args_schema, sanitized_key_mappings[t.name]) for t in tools] if tools else []
        Tracing.tracers().start_llm_trace(self, messages, current_try, anthropic_tools)
        completion = await self.client.beta.tools.messages.create(
            max_tokens=4096,
            model=self.model,
            temperature=self.temperature,
            system=messages.pop(0).content if (messages and messages[0].role == "system") else "",
            messages=await messages_to_anthropic_dicts(messages),
            tools=anthropic_tools
        )
        Tracing.tracers().end_llm_trace(completion)
        return tool_calls_to_tool_message(completion, tools, sanitized_key_mappings)

    @staticmethod
    def _output_schema_to_tool(output_schema: dict):
        output_schema = output_schema.copy()
        return Tool.create(
            name=output_schema.pop("name", None) or "record_result",
            description=output_schema.pop("description", None) or "Record your result into well-structured JSON.",
            args_schema=output_schema
        )

    @deprecated("Using a tool for now but still need to evaluate which is better")
    async def _run_output_schema_completion(self, messages: list[ChatMessage], output_schema: dict[str, Any], current_try: int = None) -> dict:
        output_schema = expand_simplified_json_schema(output_schema)
        output_prompt = Prompt.from_file(
            "output_schema",
            output_schema=json.dumps(output_schema, indent=2)
        ).format()[0].content
        if messages[0].role == "system":
            messages[0].content += f"\n\n{output_prompt}"
        else:
            messages.insert(0, SystemMessage(content=output_prompt))
        if messages[-1].role == "assistant":
            logger.warning("Ignoring trailing assistant message.")
            messages.pop()
        messages.append(AssistantMessage(content="<result>"))
        completion = await self._run_completion(messages, stop=["</result>"], current_try=current_try)
        try:
            json_object = json.loads(completion.content[0].text.strip())
        except ValueError:
            json_object = None
        return json_object

    def supports_images(self) -> bool:
        return True

    def supports_video(self) -> bool:
        return False

    def supports_audio(self) -> bool:
        return False

    def name(self) -> str:
        return "anthropic"



import inspect
from functools import wraps

from pydantic import BaseModel

from bpm_ai_core.tracing.tracing import Tracing


def trace(name: str = None, tags: list[str] = None):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            trace_name = func.__name__ if name is None else name
            params = inspect.signature(func).parameters
            inputs = {**{list(params.keys())[i]: arg for i, arg in enumerate(args) if list(params.keys())[i] != 'self'}, **kwargs}
            Tracing.tracers().start_trace(trace_name, inputs, tags or [])
            try:
                if inspect.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                if isinstance(result, dict):
                    outputs = result
                elif isinstance(result, BaseModel):
                    outputs = result.model_dump()
                else:
                    outputs = {'output': result}
                Tracing.tracers().end_trace(outputs)
                return result
            except Exception as e:
                Tracing.tracers().end_trace({'error': str(e)})
                raise e
        return wrapper
    return decorator


def span(name: str = None):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            span_name = func.__name__ if name is None else name
            params = inspect.signature(func).parameters
            inputs = {**{list(params.keys())[i]: arg for i, arg in enumerate(args) if list(params.keys())[i] != 'self'}, **kwargs}
            Tracing.tracers().start_span(span_name, inputs)
            try:
                if inspect.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                if isinstance(result, dict):
                    outputs = result
                elif isinstance(result, BaseModel):
                    outputs = result.model_dump()
                else:
                    outputs = {'output': result}
                Tracing.tracers().end_span(outputs)
                return result
            except Exception as e:
                Tracing.tracers().end_span({'error': str(e)})
                raise e
        return wrapper
    return decorator



from abc import ABC, abstractmethod

from bpm_ai_core.tracing.decorators import span


class NMTModel(ABC):
    """
    Neural Machine Translation Model
    """

    @abstractmethod
    async def _do_translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        pass

    @span(name="nmt")
    async def translate(self, text: str | list[str], target_language: str) -> str | list[str]:
        return await self._do_translate(text, target_language)



import json
import logging
import os
import re
from typing import Dict, Any, Optional, List
from urllib.parse import urljoin

from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import ChatMessage, ToolCallMessage, AssistantMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.llm.openai_chat import get_openai_client, get_azure_openai_client
from bpm_ai_core.llm.openai_chat._constants import DEFAULT_MODEL, DEFAULT_TEMPERATURE, DEFAULT_SEED, \
    DEFAULT_MAX_RETRIES, AZURE_API_KEY_ENV_VAR, OPENAI_COMPATIBLE_API_KEY_ENV_VAR
from bpm_ai_core.llm.openai_chat.util import messages_to_openai_dicts, json_schema_to_openai_function
from bpm_ai_core.tracing.tracing import Tracing

logger = logging.getLogger(__name__)

try:
    from openai import AsyncOpenAI, APIConnectionError, InternalServerError, RateLimitError, OpenAIError
    from openai.types.chat import ChatCompletionMessage as OpenAIChatCompletionMessage, ChatCompletion
    from openai.lib.azure import AsyncAzureOpenAI

    has_openai = True
except ImportError:
    has_openai = False


class ChatOpenAI(LLM):
    """
    `OpenAI` Chat large language models API.
    Also used for Azure OpenAI and other compatible servers.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.
    """

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        temperature: float = DEFAULT_TEMPERATURE,
        seed: Optional[int] = DEFAULT_SEED,
        max_retries: int = DEFAULT_MAX_RETRIES,
        client: AsyncOpenAI = None
    ):
        if not has_openai:
            raise ImportError('openai is not installed')
        if not client:
            client = get_openai_client()
        super().__init__(
            model=model,
            temperature=temperature,
            max_retries=max_retries,
            retryable_exceptions=[
                RateLimitError, InternalServerError, APIConnectionError
            ]
        )
        self.client = client
        self.seed = seed

    @classmethod
    def for_openai(
        cls,
        model: str = DEFAULT_MODEL,
        temperature: float = DEFAULT_TEMPERATURE,
        seed: Optional[int] = DEFAULT_SEED,
        max_retries: int = DEFAULT_MAX_RETRIES
    ):
        return cls(
            model=model,
            temperature=temperature,
            seed=seed,
            max_retries=max_retries
        )

    @classmethod
    def for_openai_compatible(
        cls,
        endpoint: str,
        model: str = DEFAULT_MODEL,
        temperature: float = DEFAULT_TEMPERATURE,
        seed: Optional[int] = DEFAULT_SEED,
        max_retries: int = DEFAULT_MAX_RETRIES,
        api_key_env_var: str = OPENAI_COMPATIBLE_API_KEY_ENV_VAR
    ):
        return cls(
            model=model,
            temperature=temperature,
            seed=seed,
            max_retries=max_retries,
            client=get_openai_client(
                endpoint=urljoin(endpoint, "/v1"),
                api_key=os.environ.get(api_key_env_var, "dummy")
            )
        )

    @classmethod
    def for_azure(
        cls,
        endpoint: str,
        temperature: float = DEFAULT_TEMPERATURE,
        seed: Optional[int] = DEFAULT_SEED,
        max_retries: int = DEFAULT_MAX_RETRIES
    ):
        pattern = r"(https://[^/]+)/openai/deployments/([^/]+)/[^/]+/[^/]+\?api-version=([^&]+)"
        match = re.search(pattern, endpoint)
        if match:
            azure_endpoint = match.group(1)
            deployment = match.group(2)
            api_version = match.group(3)
        else:
            raise Exception("Full endpoint with deployment and api-version required")
        api_key = os.environ.get(AZURE_API_KEY_ENV_VAR)
        if not api_key:
            raise Exception(f"API key not found ({AZURE_API_KEY_ENV_VAR} env variable)")
        return cls(
            model=deployment,
            temperature=temperature,
            seed=seed,
            max_retries=max_retries,
            client=get_azure_openai_client(
                azure_endpoint=azure_endpoint,
                api_version=api_version,
                api_key=api_key
            )
        )

    async def _generate_message(
        self,
        messages: List[ChatMessage],
        output_schema: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool]] = None,
        stop: list[str] = None,
        current_try: int = None
    ) -> AssistantMessage:
        tools = [self._output_schema_to_tool(output_schema)] if output_schema else tools
        openai_tools = [json_schema_to_openai_function(f.name, f.description, f.args_schema) for f in tools] if tools else []
        completion = await self._run_completion(messages, openai_tools, stop, current_try)
        message = completion.choices[0].message
        if message.tool_calls:
            if output_schema:
                return AssistantMessage(content=self._parse_tool_call_json(message))
            else:
                return self._openai_tool_calls_to_tool_message(message, tools)
        else:
            return AssistantMessage(content=message.content)

    async def _run_completion(
        self,
        messages: List[ChatMessage],
        tools: List[dict],
        stop: list[str] = None,
        current_try: int = None
    ) -> ChatCompletion:
        messages = await messages_to_openai_dicts(messages)
        args = {
            "model": self.model,
            "temperature": self.temperature,
            **({"seed": self.seed} if self.seed else {}),
            "messages": messages,
            **({"stop": stop} if stop else {}),
            **({
                   "tool_choice": {
                       "type": "function",
                       "function": {"name": tools[0]["function"]["name"]}
                   } if (len(tools) == 1) else ("auto" if tools else "none"),
                   "tools": tools
               } if tools else {})
        }
        Tracing.tracers().start_llm_trace(self, messages, current_try, tools)
        completion = await self.client.chat.completions.create(**args)
        Tracing.tracers().end_llm_trace(completion.choices[0].message)
        return completion

    @staticmethod
    def _output_schema_to_tool(output_schema: dict):
        output_schema = output_schema.copy()
        return Tool.create(
            name=output_schema.pop("name", None) or "store_result",
            description=output_schema.pop("description", None) or "Stores your result",
            args_schema=output_schema
        )

    @staticmethod
    def _openai_tool_calls_to_tool_message(message: OpenAIChatCompletionMessage, tools: List[Tool]) -> AssistantMessage:
        return AssistantMessage(
            name=", ".join([t.function.name for t in message.tool_calls]),
            content=message.content,
            tool_calls=[
                ToolCallMessage(
                    id=t.id,
                    name=t.function.name,
                    payload=t.function.arguments,
                    tool=next((item for item in tools if item.name == t.function.name), None)
                )
                for t in message.tool_calls
            ]
        )

    @staticmethod
    def _parse_tool_call_json(message: OpenAIChatCompletionMessage):
        try:
            json_object = json.loads(message.tool_calls[0].function.arguments)
        except ValueError as e:
            json_object = None
        return json_object

    def supports_images(self) -> bool:
        return self.model == "gpt-4-turbo"

    def supports_video(self) -> bool:
        return False

    def supports_audio(self) -> bool:
        return False

    def name(self) -> str:
        return "openai"



import logging

from typing_extensions import override

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.question_answering.question_answering import QuestionAnswering, QAResult
from bpm_ai_core.util.image import blob_as_images

try:
    from aiobotocore.session import get_session

    has_textract = True
except ImportError:
    has_textract = False

logger = logging.getLogger(__name__)

IMAGE_FORMATS = ["png", "jpeg", "tiff"]


class AmazonTextractDocVQA(QuestionAnswering):
    """

    """
    def __init__(self, region_name: str = None):
        if not has_textract:
            raise ImportError('aiobotocore is not installed')
        self.region_name = region_name

    @override
    async def _do_answer(
            self,
            context_str_or_blob: str | Blob,
            question: str
    ) -> QAResult:
        if isinstance(context_str_or_blob, str) or not (context_str_or_blob.is_image() or context_str_or_blob.is_pdf()):
            raise Exception('AmazonTextractDocVQA only supports image or PDF input')
        if context_str_or_blob.is_pdf():
            _bytes = await context_str_or_blob.as_bytes()
        else:
            _bytes = (await blob_as_images(context_str_or_blob, accept_formats=IMAGE_FORMATS, return_bytes=True))[0]

        async with get_session().create_client("textract", region_name=self.region_name) as client:
            response = await client.analyze_document(
                Document={"Bytes": _bytes},
                FeatureTypes=["QUERIES"],
                QueriesConfig={'Queries': [
                    {'Text': question}
                ]}
            )

        prediction = next((b for b in response['Blocks'] if b["BlockType"] == "QUERY_RESULT"), {})

        logger.debug(f"prediction: {prediction}")

        if prediction is None:
            raise Exception('AmazonTextractDocVQA failed to extract information.')

        return QAResult(
            answer=prediction['Text'],
            score=prediction['Confidence'] / 100.0,
            start_index=None,
            end_index=None,
        )






import asyncio

from typing_extensions import override

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.ocr.ocr import OCR, OCRResult, OCRPage
from bpm_ai_core.util.image import blob_as_images
from bpm_ai_core.util.storage import is_s3_url, parse_s3_url

try:
    from aiobotocore.session import get_session
    from textractprettyprinter.t_pretty_print import get_text_from_layout_json

    has_textract = True
except ImportError:
    has_textract = False

IMAGE_FORMATS = ["png", "jpeg", "tiff"]


class AmazonTextractOCR(OCR):
    def __init__(self, region_name: str = None):
        if not has_textract:
            raise ImportError('aiobotocore and/or amazon-textract-prettyprinter are not installed')
        self.region_name = region_name

    @override
    async def _do_process(
        self,
        blob: Blob,
        language: str = None
    ) -> OCRResult:
        if not (blob.is_pdf() or blob.is_image()):
            raise ValueError("Blob must be a PDF or an image")

        if is_s3_url(blob.path):
            pages = await self._get_pages_async(blob.path)
        else:
            pages = await self._get_pages_sync(blob)

        return OCRResult(pages=pages)

    async def _get_pages_sync(self, document: Blob):
        if document.is_pdf():
            _bytes = await document.as_bytes()
        else:
            _bytes = (await blob_as_images(document, accept_formats=IMAGE_FORMATS, return_bytes=True))[0]
        # Create a document from the image bytes asynchronously
        async with get_session().create_client("textract", region_name=self.region_name) as client:
            # Call Amazon Textract API asynchronously
            response = await client.analyze_document(
                Document={"Bytes": _bytes},
                FeatureTypes=["TABLES", "FORMS", "LAYOUT"]
            )
        # Convert Textract response to markdown using amazon-textract-prettyprinter
        markdown_pages = get_text_from_layout_json(
            textract_json=response,
            table_format="github",
            generate_markdown=True
        )
        return self.parse_pages(markdown_pages, response)

    async def _get_pages_async(self, s3_url: str):
        bucket_name, file_path = await parse_s3_url(s3_url)
        # Create a document from the image bytes asynchronously
        async with get_session().create_client("textract", region_name=self.region_name) as client:
            # Call Amazon Textract API asynchronously using start_document_analysis
            response = await client.start_document_analysis(
                DocumentLocation={'S3Object': {
                    'Bucket': bucket_name,
                    'Name': file_path
                }},
                FeatureTypes=["TABLES", "FORMS", "LAYOUT"]
            )

            # Get the job ID from the response
            job_id = response["JobId"]

            # Wait for the job to complete
            while True:
                response = await client.get_document_analysis(JobId=job_id)
                status = response["JobStatus"]
                if status in ["SUCCEEDED", "FAILED"]:
                    break
                await asyncio.sleep(1)  # Wait for 1 second before checking the status again

            if status == "FAILED":
                raise Exception(f"Document analysis failed with error: {response['StatusMessage']}")

            # Retrieve the results from the completed job
            pages = []
            markdown_pages = {}
            next_token = None
            while True:
                if next_token:
                    response = await client.get_document_analysis(JobId=job_id, NextToken=next_token)
                else:
                    response = await client.get_document_analysis(JobId=job_id)

                # Convert Textract response to markdown using amazon-textract-prettyprinter
                markdown_pages.update(get_text_from_layout_json(
                    textract_json=response,
                    table_format="github",
                    generate_markdown=True
                ))
                pages.extend(self.parse_pages(markdown_pages, response))

                next_token = response.get("NextToken")
                if not next_token:
                    break
            return pages

    @staticmethod
    def parse_pages(markdown_pages, response):
        pages = []
        page_idx = -1
        for block in response["Blocks"]:
            if block["BlockType"] == "PAGE":
                page_idx += 1
                bboxes = []
                words = []
                for _block in block["Relationships"][0]["Ids"]:
                    block_data = next(b for b in response["Blocks"] if b["Id"] == _block)
                    if block_data["BlockType"] == "LINE":
                        for word_block_id in block_data["Relationships"][0]["Ids"]:
                            word_block = next(b for b in response["Blocks"] if b["Id"] == word_block_id)
                            if word_block["BlockType"] == "WORD":
                                bbox = word_block["Geometry"]["BoundingBox"]
                                x, y, w, h = bbox["Left"], bbox["Top"], bbox["Width"], bbox["Height"]
                                bboxes.append((x, y, x + w, y + h))
                                words.append(word_block["Text"])
                page_data = OCRPage(
                    text=list(markdown_pages.values())[page_idx],
                    words=words,
                    bboxes=bboxes
                )
                pages.append(page_data)
        return pages



import asyncio
import json
import logging
import os
import pickle
from typing import Any

logger = logging.getLogger(__name__)


def remote_object_init():
    import Pyro5
    from Pyro5.api import register_class_to_dict, register_dict_to_class

    def _pickle_class_to_dict(obj: Any) -> dict:
        return {"__class__": "object", "pickle": pickle.dumps(obj)}

    def _pickle_dict_to_class(_, obj_dict: dict) -> Any:
        return pickle.loads(obj_dict["pickle"])

    register_class_to_dict(object, _pickle_class_to_dict)
    register_dict_to_class("object", _pickle_dict_to_class)

    Pyro5.config.SERIALIZER = "marshal"


def create_object_identifier(name: str, *args, **kwargs):
    args_str = '_'.join(str(arg) for arg in args)
    kwargs_str = json.dumps(kwargs, sort_keys=True).replace('"', '').replace(' ', '')
    return f"{name}:{args_str}_{kwargs_str}"


class ObjectDaemon:
    from Pyro5.server import expose

    def __init__(self, daemon, classes: list = None, instance_mode: str = None):
        self.daemon = daemon
        self.classes = classes or []
        self.instances_by_class_name = {}

        self.instance_mode = instance_mode or os.environ.get("INSTANCE_MODE", "per_class")

    @property
    def _class_dict(self):
        return {c.__name__: c for c in self.classes}

    def register_class(self, clazz):
        from Pyro5.server import expose

        def create_sync_method(async_method):
            def sync_method(self, *args, **kwargs):
                return asyncio.run(async_method(self, *args, **kwargs))

            return sync_method

        def add_sync_methods(cls):
            for attr_name in dir(cls):
                attr = getattr(cls, attr_name)
                if callable(attr) and asyncio.iscoroutinefunction(attr):
                    sync_method_name = f"{attr_name}_sync"
                    sync_method = create_sync_method(attr)
                    setattr(cls, sync_method_name, sync_method)
            return cls

        add_sync_methods(clazz)
        expose(clazz)
        self.classes.append(clazz)

    @expose
    def create_instance(self, class_name: str, *args, **kwargs) -> str:
        object_id = create_object_identifier(class_name, *args, **kwargs)

        if object_id in self.daemon.objectsById:
            return object_id
        else:
            if self.instance_mode == "per_call":
                for obj in self.instances_by_class_name:
                    logging.info(f"Deleting old object for class {class_name}")
                    self.daemon.unregister(obj)
                    del obj
            elif self.instance_mode == "per_class" and class_name in self.instances_by_class_name:
                obj = self.instances_by_class_name[class_name]
                logging.info(f"Deleting old object for class {class_name}")
                self.daemon.unregister(obj)
                del obj
            else:  # per_params
                pass

            clazz = self._class_dict[class_name]
            instance = clazz(*args, **kwargs)
            self.daemon.register(instance, objectId=object_id)
            self.instances_by_class_name[class_name] = instance

            logging.info(f"Created new object for class {class_name}")

            return object_id

    def serve(self):
        self.daemon.requestLoop()


def create_remote_object_daemon(host: str = "localhost", port: int = 1337) -> ObjectDaemon:
    import Pyro5.api
    remote_object_init()
    pyro_daemon = Pyro5.api.Daemon(host=host, port=port)
    daemon = ObjectDaemon(pyro_daemon)
    pyro_daemon.register(daemon, objectId="__object_creator__")
    logging.info(f"Object daemon created for {host}:{port}")
    return daemon


def remote_object(name: str, host: str = "localhost", port: int = 6666, *args, **kwargs) -> Any:
    import Pyro5.errors
    remote_object_init()
    try:
        object_id = create_object_identifier(name, *args, **kwargs)
        return _get_proxy(object_id, host, port, make_async=True)
    except Pyro5.errors.CommunicationError:
        logging.info(f"Remote object for {name} not found, creating a new one...")
        creator = _get_proxy("__object_creator__", host, port)
        object_id = creator.create_instance(name, *args, **kwargs)
        return _get_proxy(object_id, host, port, make_async=True)


def _get_proxy(object_id, host, port, make_async=False):
    import Pyro5.api
    from Pyro5.client import Proxy

    class AsyncProxy(Pyro5.api.Proxy):
        def __getattr__(self, name):
            attr = super().__getattr__(name + "_sync")
            return self._sync_to_async_wrapper(attr)

        def _sync_to_async_wrapper(self, sync_func):
            s = super()

            def ownership_wrapper(f, *args, **kwargs):
                # claim ownership for object as we are running in a new thread
                s._pyroClaimOwnership()
                return f(*args, **kwargs)

            async def async_wrapper(*args, **kwargs):
                coro = asyncio.to_thread(ownership_wrapper, sync_func, *args, **kwargs)
                return await coro

            return async_wrapper

    ProxyClass = AsyncProxy if make_async else Proxy
    proxy = ProxyClass(f"PYRO:{object_id}@{host}:{port}")
    proxy._pyroBind()
    return proxy



import hashlib
import logging
from contextvars import ContextVar

from bpm_ai_core.llm.anthropic_chat._constants import default_client_kwargs

logger = logging.getLogger(__name__)

try:
    from anthropic import AsyncAnthropic

    _clients: ContextVar[dict[str, AsyncAnthropic]] = ContextVar('anthropic_clients', default={})
except ImportError:
    pass


def get_anthropic_client(endpoint: str = None, api_key: str = None) -> AsyncAnthropic:
    client_map = _clients.get()
    hash_key = hashlib.sha256(((endpoint or "default") + (api_key or "default")).encode()).hexdigest()
    if hash_key in client_map.keys():
        return client_map[hash_key]
    else:
        client = AsyncAnthropic(
            base_url=endpoint,
            api_key=api_key,
            **default_client_kwargs
        )
        client_map[hash_key] = client
        return client



import io
import os

import requests

audio_ext_map = {
    'flac': 'audio/flac',
    'mp3': 'audio/mpeg',
    'mp4': 'audio/mpeg',
    'mpeg': 'audio/mpeg',
    'mpga': 'audio/mpeg',
    'm4a': 'audio/mpeg',
    'ogg': 'audio/ogg',
    'wav': 'audio/vnd.wav',
    'webm': 'audio/webm',
}


def load_audio(path: str) -> io.BytesIO:
    """
    Load an audio file from a local path or a web URL into a BytesIO object.

    Parameters:
    - path (str): A file system path or a URL of an audio file.

    Returns:
    - BytesIO: A BytesIO object containing the audio data.
    """
    if path.startswith('http://') or path.startswith('https://'):
        # Handle web URL
        response = requests.get(path)
        audio = io.BytesIO(response.content)
        if path.endswith(tuple(audio_ext_map.keys())):
            audio.name = f"audio.{path.rsplit('.', 1)[-1]}"
    elif os.path.isfile(path):
        # Handle local file path
        with open(path, 'rb') as f:
            audio = io.BytesIO(f.read())
            audio.name = f.name
    else:
        raise ValueError("The path provided is neither a valid URL nor a file path.")

    return audio


import asyncio
from typing import List

from typing_extensions import override

from bpm_ai_core.translation.nmt import NMTModel

try:
    from aiobotocore.session import get_session

    has_amazon_translate = True
except ImportError:
    has_amazon_translate = False


class AmazonTranslate(NMTModel):
    """Amazon Translate NMT Model"""
    def __init__(self, region_name: str = None):
        if not has_amazon_translate:
            raise ImportError("aiobotocore is not installed")
        self.region_name = region_name

    async def _translate_single(self, text: str, target_language: str) -> str:
        async with get_session().create_client('translate', region_name=self.region_name) as client:
            response = await client.translate_text(
                Text=text,
                SourceLanguageCode='auto',
                TargetLanguageCode=target_language
            )
            return response['TranslatedText']

    @override
    async def _do_translate(self, text: str | List[str], target_language: str) -> str | List[str]:
        if isinstance(text, str):
            return await self._translate_single(text, target_language)
        else:
            tasks = [self._translate_single(t, target_language) for t in text]
            return await asyncio.gather(*tasks)



import asyncio
import inspect
import json
from typing import Optional, Literal, Any, Union, List

from pydantic import BaseModel, Field, ConfigDict

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.tracing.tracing import Tracing


class ChatMessage(BaseModel):
    content: Optional[Union[str, dict, List[Union[str, Blob]]]] = None
    """
    The contents of the message. 
    Either a string for normal completions, 
    or a list of strings and blobs for multimodal completions, 
    or a dict for prediction with output schema.
    """

    role: Literal["system", "user", "assistant", "tool"]
    """
    The role of the messages author.
    One of `system`, `user`, `assistant`, or `tool`.
    """

    name: Optional[str] = None
    """
    The name of the author of this message.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)


class SystemMessage(ChatMessage):

    role: str = Field("system")


class UserMessage(ChatMessage):

    role: str = Field("user")


class ToolCallMessage(BaseModel):

    id: str

    type: str = Field("function")

    name: str
    """
    The name of the tool that was used.
    """

    payload: Any

    tool: Optional[Tool] = None

    def payload_dict(self) -> dict:
        if isinstance(self.payload, dict):
            return self.payload
        elif isinstance(self.payload, str):
            try:
                return json.loads(self.payload)
            except ValueError as e:
                raise Exception(f"Payload could not be converted to a dict: {e}")
        else:
            raise Exception("Payload has unexpected type.")

    async def run_tool_function(self) -> Any:
        _callable = self.tool.callable
        inputs = self.payload_dict()
        Tracing.tracers().start_tool_trace(self.tool, inputs)
        if inspect.iscoroutinefunction(_callable):
            result = await _callable(**inputs)
        else:
            result = _callable(**inputs)
        Tracing.tracers().end_tool_trace(result)
        return result


class ToolResultMessage(ChatMessage):

    role: str = Field("tool")

    id: str
    """
    The id of the tool.
    """


class AssistantMessage(ChatMessage):

    role: str = Field("assistant")

    tool_calls: List[ToolCallMessage] | None = None

    def has_tool_calls(self) -> bool:
        return self.tool_calls and len(self.tool_calls) > 0

    async def run_all_tool_functions(self) -> List[Any]:
        return [await t.run_tool_function() for t in self.tool_calls]

    async def run_all_tool_functions_parallel(self) -> List[Any]:
        return await asyncio.gather(*[t.run_tool_function() for t in self.tool_calls])



import os
from typing import List
from urllib.parse import urlparse

from bpm_ai_core.util.audio import audio_ext_map
from bpm_ai_core.util.image import image_ext_map, pdf_ext_map
from bpm_ai_core.util.text import text_ext_map

supported_ext_map = {**audio_ext_map, **image_ext_map, **pdf_ext_map, **text_ext_map}
supported_extensions = supported_ext_map.keys()


def guess_mimetype(filename: str) -> str | None:
    if is_supported_file(filename, list(image_ext_map.keys())):
        return image_ext_map[_extract_extension(filename)]
    elif is_supported_file(filename, list(text_ext_map.keys())):
        return text_ext_map[_extract_extension(filename)]
    elif is_supported_file(filename, list(pdf_ext_map.keys())):
        return "application/pdf"
    elif is_supported_file(filename, list(audio_ext_map.keys())):
        return audio_ext_map[_extract_extension(filename)]


def is_supported_file(url_or_path: str, extensions: List[str] = supported_extensions) -> bool:
    file_extension = _extract_extension(url_or_path)
    # Normalize the extensions to lowercase
    extensions = [ext.lower() for ext in extensions]
    # Check if the file extension is in the list of supported extensions
    return file_extension in extensions


def is_supported_img_file(url_or_path: str) -> bool:
    return is_supported_file(url_or_path, extensions=list(image_ext_map.keys()) + list(pdf_ext_map.keys()))


def is_supported_audio_file(url_or_path: str) -> bool:
    return is_supported_file(url_or_path, extensions=list(audio_ext_map.keys()))


def is_supported_text_file(url_or_path: str) -> bool:
    return is_supported_file(url_or_path, extensions=list(text_ext_map.keys()))


def _extract_extension(url_or_path):
    url_or_path = url_or_path.strip()
    # Extract the path from URL if it's a URL
    parsed_url = urlparse(url_or_path)
    path = parsed_url.path if parsed_url.scheme else url_or_path
    # Remove trailing slash if present
    if path.endswith('/'):
        path = path[:-1]
    # Extract the file extension
    _, file_extension = os.path.splitext(path)
    file_extension = file_extension.lower().lstrip('.')
    return file_extension


def is_file(s: str) -> bool:
    return s.startswith('http://') or s.startswith('https://') or os.path.exists(os.path.dirname(s))



from typing import List, Any

from bpm_ai_core.llm.common.llm import LLM
from bpm_ai_core.llm.common.message import ChatMessage, ToolCallMessage, AssistantMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.tracing.tracing import Tracing


def messages_to_str(messages: List[ChatMessage]) -> str:
    return "\n\n".join([m.content for m in messages])


def tool_response(name, payload):
    return AssistantMessage(tool_calls=[ToolCallMessage(id="fake", name=name, payload=payload)])


class FakeLLM(LLM):
    """Fake LLM for testing purposes."""

    def __init__(
        self,
        responses: list[AssistantMessage] | None = None,
        tools: list[list[Tool]] | None = None,
        supports_images: bool = False,
        supports_video: bool = False,
        supports_audio: bool = False,
        name: str = "test-llm"
    ):
        super().__init__("test-model")
        self._name = name

        self.responses = responses or []
        self.tools = tools or []
        self.response_idx: int = 0
        self.requests = []

        self._supports_images = supports_images
        self._supports_video = supports_video
        self._supports_audio = supports_audio

    async def _generate_message(
        self, messages: list[ChatMessage],
        output_schema: dict[str, Any] | None = None,
        tools: list[Tool] | None = None,
        stop: list[str] = None,
        current_try: int = None
    ) -> AssistantMessage:
        self.requests += [messages]
        self.tools += [tools]

        Tracing.tracers().start_llm_trace(self, messages, current_try, tools)

        if self.responses:
            response = self.responses[self.response_idx]
            if response.has_tool_calls():
                for c in response.tool_calls:
                    c.tool = next((item for item in tools if item.name == c.name), None)
            self.response_idx += 1
        else:
            response = None

        Tracing.tracers().end_llm_trace(response)

        return response

    def assert_last_request_contains(self, text: str):
        assert text in messages_to_str(self.requests[-1])

    def assert_last_request_not_contains(self, text: str):
        assert text not in messages_to_str(self.requests[-1])

    def assert_no_request(self):
        assert len(self.requests) == 0

    def assert_last_request_defined_tool(self, tool_name: str, is_fixed_tool_choice: bool = False):
        assert tool_name in [f.name for f in self.tools[-1]]
        if is_fixed_tool_choice:
            assert len(self.tools[-1]) == 1 and self.tools[-1][0].name == tool_name

    def supports_images(self) -> bool:
        return self._supports_images

    def supports_video(self) -> bool:
        return self._supports_video

    def supports_audio(self) -> bool:
        return self._supports_audio

    def name(self) -> str:
        return self._name



import logging
from typing import List, Dict, Any

from PIL.Image import Image

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.llm.common.message import ChatMessage, ToolResultMessage, AssistantMessage
from bpm_ai_core.util.image import base64_encode_image, blob_as_images

logger = logging.getLogger(__name__)

ACCEPTED_IMAGE_FORMATS = ["jpeg", "png", "gif", "webp"]


def get_openai_tool_call_dict(message: AssistantMessage):
    return {
        "tool_calls": [
            {
                "type": "function",
                "id": t.id,
                "function": {
                    "name": t.name,
                    "arguments": t.payload
                }
            }
            for t in message.tool_calls
        ]
    }


async def messages_to_openai_dicts(messages: List[ChatMessage]):
    return [await message_to_openai_dict(m) for m in messages]


async def message_to_openai_dict(message: ChatMessage) -> dict:
    if isinstance(message, AssistantMessage) and message.has_tool_calls():
        extra_dict = {
            **get_openai_tool_call_dict(message)
        }
    elif isinstance(message, ToolResultMessage):
        extra_dict = {
            "tool_call_id": message.id
        }
    else:
        extra_dict = {}

    if isinstance(message.content, str):
        content = message.content
    elif isinstance(message.content, list):
        content = []
        for e in message.content:
            if isinstance(e, str):
                content.append(str_to_openai_text_dict(e))
            elif isinstance(e, Blob) and (e.is_image() or e.is_pdf()):
                images = await blob_as_images(e, accept_formats=ACCEPTED_IMAGE_FORMATS)
                for i, image in enumerate(images):
                    content.append(str_to_openai_text_dict(f"Image / Page {i + 1}:"))
                    content.append(image_to_openai_image_dict(image))
            elif isinstance(e, Blob) and (e.is_text()):
                text = (await e.as_bytes()).decode("utf-8")
                filename = (" name='" + e.path + "'") if e.path else ''
                text = f"<file{filename}>\n{text}\n</file>"
                content.append(str_to_openai_text_dict(text))
            else:
                raise ValueError(
                    "Elements in ChatMessage.content must be str or Blob (image/pdf/text)"
                )
    else:
        content = None
        logger.warning(
            "ChatMessage.content must be of type str or List[Union[str, PIL.Image]] if used for chat completions."
        )
    return {
        "role": message.role,
        **({"content": content} if content else {}),
        **extra_dict,
        **({"name": message.name} if message.name else {})
    }


def image_to_openai_image_dict(image: Image) -> dict:
    return {
        "type": "image_url",
        "image_url": {
            "url": f"data:image/{image.format.lower()};base64,{base64_encode_image(image)}"
        }
    }


def str_to_openai_text_dict(text: str) -> dict:
    return {
        "type": "text",
        "text": text
    }


def json_schema_to_openai_function(name: str, desc: str, schema: Dict[str, Any]) -> dict:
    return {
        "type": "function",
        "function": {
            "name": name,
            "description": desc,
            "parameters": schema
        }
    }



import json
import logging
import re
from typing import List, Any, Callable

from PIL.Image import Image

from bpm_ai_core.llm.common.blob import Blob
from bpm_ai_core.llm.common.message import ChatMessage, ToolResultMessage, AssistantMessage, ToolCallMessage
from bpm_ai_core.llm.common.tool import Tool
from bpm_ai_core.util.image import base64_encode_image, blob_as_images
from bpm_ai_core.util.json_schema import sanitize_schema, desanitize_json
from bpm_ai_core.util.linguistics import replace_diacritics

logger = logging.getLogger(__name__)

ACCEPTED_IMAGE_FORMATS = ["jpeg", "png", "gif", "webp"]


async def messages_to_anthropic_dicts(messages: List[ChatMessage]):
    return [await message_to_anthropic_dict(m) for m in messages]


async def message_to_anthropic_dict(message: ChatMessage) -> dict:
    if isinstance(message, AssistantMessage) and message.has_tool_calls():
        return tool_calls_message_to_anthropic_dict(message)
    elif isinstance(message, ToolResultMessage):
        return tool_result_message_to_anthropic_dict(message)
    elif isinstance(message.content, str):
        content = message.content
    elif isinstance(message.content, list):
        content = []
        for e in message.content:
            if isinstance(e, str):
                content.append(str_to_anthropic_text_dict(e))
            elif isinstance(e, Blob) and (e.is_image() or e.is_pdf()):
                images = await blob_as_images(e, accept_formats=ACCEPTED_IMAGE_FORMATS)
                for i, image in enumerate(images):
                    content.append(str_to_anthropic_text_dict(f"Image / Page {i+1}:"))
                    content.append(image_to_anthropic_image_dict(image))
            elif isinstance(e, Blob) and (e.is_text()):
                text = (await e.as_bytes()).decode("utf-8")
                filename = (" name='" + e.path + "'") if e.path else ''
                text = f"<file{filename}>\n{text}\n</file>"
                content.append(str_to_anthropic_text_dict(text))
            else:
                raise ValueError(
                    "Elements in ChatMessage.content must be str or Blob (image/pdf/text)"
                )
    else:
        content = None
        logger.warning(
            "ChatMessage.content must be of type str or List[Union[str, PIL.Image]] if used for chat completions."
        )
    return {
        "role": message.role,
        **({"content": content} if content else {}),
        **({"name": message.name} if message.name else {})
    }


def tool_calls_message_to_anthropic_dict(message: AssistantMessage) -> dict:
    return {
        "role": "assistant",
        "content": ([{"type": "text", "text": message.content}] if message.content else [])
                 + [{"type": "tool_use", "id": call.id, "name": call.name, "input": call.payload_dict()}
                    for call in message.tool_calls]
    }


def tool_result_message_to_anthropic_dict(message: ToolResultMessage, is_error: bool = False) -> dict:
    return {
      "role": "user",
      "content": [
        {
          "type": "tool_result",
          "tool_use_id": message.id,
          "content": message.content if isinstance(message.content, str) else json.dumps(message.content, indent=2),
          **({"is_error": True} if is_error else {})
        }
      ]
    }


def image_to_anthropic_image_dict(image: Image) -> dict:
    return {
        "type": "image",
        "source": {
            "type": "base64",
            "media_type": f"image/{image.format.lower()}",
            "data": base64_encode_image(image),
        }
    }


def str_to_anthropic_text_dict(text: str) -> dict:
    return {
        "type": "text",
        "text": text
    }


def json_schema_to_anthropic_tool(name: str, desc: str, schema: dict[str, Any], key_mapping: dict) -> dict:
    return {
        "name": name,
        "description": desc,
        "input_schema": sanitize_schema(schema, sanitize_key, key_mapping)
    }


def sanitize_key(item):
    return re.sub(r'[^a-zA-Z0-9_-]+', '', replace_diacritics(item).replace(' ', '_'))[:64]


def tool_calls_to_tool_message(message, tools: List[Tool], key_mappings: dict) -> AssistantMessage:
    from anthropic.types import TextBlock
    from anthropic.types.beta.tools import ToolUseBlock

    texts = [c.text for c in message.content if isinstance(c, TextBlock)]
    tool_uses = [c for c in message.content if isinstance(c, ToolUseBlock)]
    return AssistantMessage(
        name=", ".join([t.name for t in tool_uses]),
        content="\n\n".join(texts),
        tool_calls=[
            ToolCallMessage(
                id=t.id,
                name=t.name,
                payload=desanitize_json(t.input, key_mappings[t.name]),
                tool=next((item for item in tools if item.name == t.name), None)
            )
            for t in tool_uses
        ]
    )



import json


def dict_to_md(json_obj, depth=2):
    markdown = ""
    if isinstance(json_obj, dict):
        for key, value in json_obj.items():
            if isinstance(value, list) and all(isinstance(item, dict) for item in value):
                # If it's a list of objects, pretty print the JSON representation
                markdown += f"{'#' * depth} {key}\n\n```json\n{json.dumps(value, indent=2, ensure_ascii=False)}\n```\n\n"
            elif isinstance(value, (dict, list)):
                markdown += f"{'#' * depth} {key}\n"
                markdown += dict_to_md(value, depth + 1)
            else:
                markdown += f"{key}: {value}\n"
        markdown += "\n"  # One newline at the end of an object
    elif isinstance(json_obj, list):
        for item in json_obj:
            if isinstance(item, (dict, list)):
                markdown += dict_to_md(item, depth)
            else:
                markdown += f"- {item}\n"
        markdown += "\n"  # One newline at the end of a list
    else:
        markdown += f"{json_obj}\n\n"  # If any other type, print it with a newline
    return markdown


import httpx

DEFAULT_MODEL = "gpt-4-turbo"
DEFAULT_TEMPERATURE = 0.0
DEFAULT_SEED = 42
DEFAULT_MAX_RETRIES = 8
OPENAI_COMPATIBLE_API_KEY_ENV_VAR = "LLM_API_KEY"
AZURE_API_KEY_ENV_VAR = "AZURE_OPENAI_API_KEY"

default_client_kwargs = {
    "http_client": httpx.AsyncClient(
        limits=httpx.Limits(max_connections=1000, max_keepalive_connections=100)
    ),
    "timeout": 600,
    "max_retries": 0  # we use own retry logic
}


from abc import ABC, abstractmethod

from pydantic import BaseModel


class TokenSpan(BaseModel):
    label: str
    score: float
    word: str
    start: int
    end: int


class TokenClassificationResult(BaseModel):
    spans: list[TokenSpan]


class ZeroShotTokenClassifier(ABC):
    """
    Zero Shot Token Classification Model
    """

    @abstractmethod
    async def _do_classify(
            self,
            text: str,
            classes: list[str],
            confidence_threshold: float | None = None
    ) -> TokenClassificationResult:
        pass

    async def classify(
            self,
            text: str,
            classes: list[str],
            confidence_threshold: float | None = None
    ) -> TokenClassificationResult:
        return await self._do_classify(
            text=text,
            classes=classes,
            confidence_threshold=confidence_threshold
        )



import logging

from bpm_ai_core.tracing.tracer import Tracer
from bpm_ai_core.util.image import replace_base64_values

try:
    from langfuse import Langfuse
    from langfuse.client import StatefulTraceClient, StatefulSpanClient, StatefulGenerationClient
    has_langfuse = True
except ImportError:
    has_langfuse = False

logger = logging.getLogger(__name__)


class LangfuseTracer(Tracer):
    """
    `Langfuse` Tracer (https://langfuse.com)

    To use, you should have the ``langfuse`` python package installed, and the
    environment variables ``LANGFUSE_SECRET_KEY`` and ``LANGFUSE_PUBLIC_KEY`` set.
    """

    def __init__(self):
        if not has_langfuse:
            raise ImportError('langfuse is not installed')
        self.langfuse = Langfuse()

        self.trace: StatefulTraceClient | None = None
        self.implicit_trace: bool = False
        self.span_stack: list[StatefulSpanClient] = []
        self.generation: StatefulGenerationClient | None = None
        self.current_tool = None

    def start_trace(self, name: str, inputs: dict | list[dict], tags: list[str] = None):
        if self.trace:
            raise Exception("Trace already started for this thread - end it first")
        self.trace = self.langfuse.trace(
            name=name,
            input=inputs,
            tags=tags
        )

    def end_trace(self, outputs: dict, error_msg: str = None):
        if not self.trace:
            raise Exception("No trace started for this thread")
        self.trace.update(
            output=outputs,
            level="ERROR" if error_msg else None,
            status_message=error_msg
        )
        logger.info(f"[Langfuse Trace Finished] {self.trace.get_trace_url()}")
        self.trace = None
        self.span_stack = []

    def start_span(self, name: str, inputs: dict = None):
        if not self.trace and not self.span_stack:
            self.start_trace("unnamed", inputs=inputs)
            self.implicit_trace = True
            client = self.trace
        elif self.span_stack:
            client = self.span_stack[-1]
        else:
            client = self.trace
        span = client.span(name=name, input=inputs)
        self.span_stack.append(span)

    def end_span(self, output=None, error_msg: str = None):
        if not self.span_stack:
            raise Exception("No span started for this thread")
        span = self.span_stack.pop()
        span.end(
            output=output,
            level="ERROR" if error_msg else None,
            status_message=error_msg
        )
        if self.implicit_trace:
            self.end_trace(outputs=output)

    def start_llm_trace(self, llm, messages: list, current_try: int, tools=None):
        messages = replace_base64_values(messages)
        if self.span_stack:
            client = self.span_stack[-1]
        else:
            client = self.trace

        self.generation = client.generation(
            model=llm.model,
            model_parameters={
                "temperature": llm.temperature
            },
            input=messages,
            metadata={
                "current_try": current_try,
                "max_tries": llm.max_retries + 1,
                "tools": tools
            }
        )

    def end_llm_trace(self, completion=None, error_msg: str = None):
        if not self.generation:
            raise Exception("No generation started for this thread")
        self.generation.end(
            output=completion,
            level="ERROR" if error_msg else None,
            status_message=error_msg
        )

    def start_tool_trace(self, tool, inputs: dict):
        self.start_span(tool.name, inputs)

    def end_tool_trace(self, outputs: dict | None = None, error_msg: str | None = None):
        self.end_span(outputs, error_msg)

    def event(self, name: str, inputs: dict | None = None, outputs: dict | None = None, error_msg: str | None = None):
        if not self.trace:
            raise Exception("No trace started for this thread")
        if self.span_stack:
            client = self.span_stack[-1]
        else:
            client = self.trace
        client.event(
            name=name,
            inputs=inputs,
            output=outputs,
            level="ERROR" if error_msg else None,
            status_message=error_msg
        )

    def finalize(self):
        self.langfuse.flush()




text_ext_map = {
    'json': 'application/json',
    'js': 'application/javascript',
    'xml': 'application/xml',
    'sh': 'application/x-sh',
    'py': 'application/x-python',
    'php': 'application/x-httpd-php',
    'phtml': 'application/x-httpd-php',
    'latex': 'application/x-latex',
    'tex': 'application/x-tex',
    'htm': 'text/html',
    'html': 'text/html',
    'shtml': 'text/html',
    'talk': 'text/x-speech',
    'spc': 'text/x-speech',
    'rtf': 'text/rtf',
    'rtx': 'text/richtext',
    'txt': 'text/plain',
    'md': 'text/markdown',
    'markdown': 'text/markdown',
    'css': 'text/css',
    'csv': 'text/csv',
    'ics': 'text/calendar',
}



import logging

from bpm_ai_core.tracing.tracer import Tracer
from bpm_ai_core.util.image import replace_base64_values

logger = logging.getLogger("logging-tracer")
logger.setLevel(logging.INFO)


class LoggingTracer(Tracer):

    def __init__(self):
        self._level = 0

    def _indent(self):
        return ("|--" if self._level == 0 else "|") + ("  " * self._level) + ("|--" if self._level > 0 else "")

    def start_trace(self, name: str, inputs: dict, tags: list[str] = None):
        logger.info(f"[TRACE START] {name},{f'{tags}, ' if tags else ''} inputs={inputs}")

    def end_trace(self, outputs, error_msg: str = None):
        if error_msg:
            logger.error(f"[TRACE ERROR] {error_msg}")
        else:
            logger.info(f"[TRACE END] outputs={outputs}")

    def start_span(self, name: str, inputs: dict):
        logger.info(self._indent() + f"[SPAN START] {name}, inputs={inputs}")
        self._level += 1

    def end_span(self, outputs: dict, error_msg: str = None):
        self._level -= 1
        if error_msg:
            logger.error(f"[SPAN ERROR] {error_msg}")
        else:
            logger.info(self._indent() + f"[SPAN END] outputs={outputs}")

    def event(self, name: str, inputs: dict = None, outputs: dict = None, error_msg: str = None):
        if error_msg:
            logger.error(self._indent() + f"[EVENT] {error_msg}")
        else:
            logger.info(self._indent() + f"[EVENT] {name}, inputs={inputs}, outputs={outputs}")

    def start_llm_trace(self, llm, messages, current_try, tools=None):
        logger.info(self._indent() + f"[LLM <] {llm.model}, current_try: {current_try}, tools={tools}, messages={replace_base64_values(messages)}")

    def end_llm_trace(self, completion=None, error_msg=None):
        if error_msg:
            logger.error(self._indent() + f"[LLM COMPLETION ERROR] {error_msg}")
            return
        logger.info(self._indent() + f"[LLM >] {completion}")

    def start_tool_trace(self, tool, inputs):
        logger.info(self._indent() + f"[TOOL] {tool.name}({inputs})")

    def end_tool_trace(self, output=None, error_msg=None):
        if error_msg:
            logger.error(self._indent() + f"[TOOL ERROR] {error_msg}")
        else:
            logger.info(self._indent() + f"[TOOL RESULT] {output}")

    def finalize(self):
        pass



import os
from io import BytesIO
from pathlib import PurePath
from typing import Union, Optional, Dict, Any, Self, cast

import requests
from pydantic import BaseModel, Field, model_validator

from bpm_ai_core.util.file import guess_mimetype
from bpm_ai_core.util.storage import read_file_from_azure_blob, read_file_from_s3, is_s3_url, is_azure_blob_url


class Blob(BaseModel):
    """Blob represents raw data by either reference or value.

    Provides an interface to materialize the blob in different representations.

    Based on:
    https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/document_loaders/blob_loaders.py
    """

    data: Union[bytes, str, None]
    """Raw data associated with the blob."""

    mimetype: Optional[str] = None
    """MimeType not to be confused with a file extension."""

    path: Optional[Union[str, PurePath]] = None
    """Location where the original content was found."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata about the blob (e.g., source)"""

    class Config:
        arbitrary_types_allowed = True
        frozen = True

    @property
    def source(self) -> Optional[str]:
        """The source location of the blob as string if known otherwise none.

        If a path is associated with the blob, it will default to the path location.

        Unless explicitly set via a metadata field called "source", in which
        case that value will be used instead.
        """
        if self.metadata and "source" in self.metadata:
            return cast(Optional[str], self.metadata["source"])
        return str(self.path) if self.path else None

    @model_validator(mode='after')
    def check_blob_is_valid(self) -> Self:
        """Verify that either data or path is provided."""
        if not self.data and not self.path:
            raise ValueError("Either data or path must be provided")
        return self

    def is_image(self) -> bool:
        return self.mimetype.startswith("image/") if self.mimetype else False

    def is_pdf(self) -> bool:
        return self.mimetype == "application/pdf" if self.mimetype else False

    def is_audio(self) -> bool:
        return self.mimetype.startswith("audio/") if self.mimetype else False

    def is_video(self) -> bool:
        return self.mimetype.startswith("video/") if self.mimetype else False

    def is_text(self) -> bool:
        app_text_mimetypes = [
            'application/json',
            'application/javascript',
            'application/xml',
            'application/x-sh',
            'application/x-python',
            'application/x-httpd-php',
            'application/x-httpd-php',
            'application/x-latex',
            'application/x-tex',
        ]
        return (self.mimetype.startswith("text/") or self.mimetype in app_text_mimetypes) if self.mimetype else False

    async def as_bytes(self) -> bytes:
        """Read data as bytes."""
        if self.data is None and (self.path.startswith('http://') or self.path.startswith('https://')):
            response = requests.get(self.path)
            return response.content
        elif self.data is None and is_s3_url(self.path):
            return await read_file_from_s3(self.path)
        elif self.data is None and is_azure_blob_url(self.path):
            return await read_file_from_azure_blob(self.path)
        elif isinstance(self.data, bytes):
            return self.data
        elif isinstance(self.data, str):
            return self.data.encode("utf-8")
        elif self.data is None and self.path:
            with open(str(self.path), "rb") as f:
                return f.read()
        else:
            raise ValueError(f"Unable to get bytes for blob {self}")

    async def as_bytes_io(self) -> BytesIO:
        return BytesIO(await self.as_bytes())

    @classmethod
    def from_path_or_url(
            cls,
            path: Union[str, PurePath],
            *,
            mime_type: Optional[str] = None,
            guess_type: bool = True,
            metadata: Optional[dict] = None,
    ) -> "Blob":
        """Load the blob from a path like object.

        Args:
            path: path like object to file to be read
            mime_type: if provided, will be set as the mime-type of the data
            guess_type: If True, the mimetype will be guessed from the file extension,
                        if a mime-type was not provided
            metadata: Metadata to associate with the blob

        Returns:
            Blob instance
        """
        if mime_type is None and guess_type:
            _mimetype = guess_mimetype(path) if guess_type else None
        else:
            _mimetype = mime_type

        # Convert a path to an absolute path
        if os.path.isfile(path):
            path = os.path.abspath(path)

        # We do not load the data immediately, instead we treat the blob as a
        # reference to the underlying data.
        return cls(
            data=None,
            mimetype=_mimetype,
            path=path,
            metadata=metadata if metadata is not None else {},
        )

    @classmethod
    def from_data(
            cls,
            data: Union[str, bytes],
            *,
            mime_type: str,
            path: Optional[str] = None,
            metadata: Optional[dict] = None,
    ) -> "Blob":
        """Initialize the blob from in-memory data.

        Args:
            data: the in-memory data associated with the blob
            mime_type: if provided, will be set as the mime-type of the data
            path: if provided, will be set as the source from which the data came
            metadata: Metadata to associate with the blob

        Returns:
            Blob instance
        """
        return cls(
            data=data,
            mimetype=mime_type,
            path=path,
            metadata=metadata if metadata is not None else {},
        )

    def __repr__(self) -> str:
        """Define the blob representation."""
        str_repr = f"Blob {id(self)}"
        if self.source:
            str_repr += f" {self.source}"
        return str_repr


import os
from contextvars import ContextVar

from bpm_ai_core.tracing.delegate import DelegateTracer
from bpm_ai_core.tracing.tracer import Tracer

_tracers: ContextVar[list[Tracer] | None] = ContextVar('tracers', default=None)


def _configure_tracers():
    tracers = []
    if not os.environ.get("BPM_AI_CONSOLE_TRACING_DISABLED", False):
        from bpm_ai_core.tracing.logging import LoggingTracer
        tracers.append(LoggingTracer())
    if os.environ.get("LANGFUSE_SECRET_KEY"):
        from bpm_ai_core.tracing.langfuse import LangfuseTracer
        tracers.append(LangfuseTracer())
    _tracers.set(tracers)


class Tracing:
    @staticmethod
    def tracers() -> Tracer:
        if _tracers.get() is None:
            _configure_tracers()
        return DelegateTracer(_tracers.get())

    @staticmethod
    def finalize():
        Tracing.tracers().finalize()
